from __future__ import annotations

import h5py
import numpy as np
import nxmx
import pytest
import scipy.stats as stats

from cctbx import factor_ev_angstrom
from scitbx.array_family import flex

import dxtbx.model
import dxtbx.nexus


def test_get_dxtbx_goniometer_multi_axis(nxmx_example):
    sample = nxmx.NXmx(nxmx_example).entries[0].samples[0]
    gonio = dxtbx.nexus.get_dxtbx_goniometer(sample)
    assert isinstance(gonio, dxtbx.model.MultiAxisGoniometer)
    assert gonio.get_rotation_axis() == (1.0, 0.0, 0.0)
    assert list(gonio.get_angles()) == [0.0, 0.0, 0.0]
    assert list(gonio.get_axes()) == [
        (1.0, 0.0, 0.0),
        (0.0, 0.0, -1.0),
        (1.0, 0.0, 0.0),
    ]
    assert list(gonio.get_names()) == ["phi", "chi", "omega"]
    assert gonio.get_scan_axis() == 2


@pytest.fixture
def nxsample_single_axis():
    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        entry = f.create_group("/entry")
        entry.attrs["NX_class"] = "NXentry"
        entry["definition"] = "NXmx"

        sample = entry.create_group("sample")
        sample.attrs["NX_class"] = "NXsample"
        sample["name"] = "mysample"
        sample["depends_on"] = b"/entry/sample/transformations/omega"

        transformations = sample.create_group("transformations")
        transformations.attrs["NX_class"] = "NXtransformations"
        omega = transformations.create_dataset("omega", data=np.arange(0, 60, 0.1))
        omega.attrs["depends_on"] = b"."
        omega.attrs["transformation_type"] = b"rotation"
        omega.attrs["units"] = b"deg"
        omega.attrs["vector"] = np.array([0, 1, 0])

        yield f


def test_get_dxtbx_goniometer_single_axis(nxsample_single_axis):
    sample = nxmx.NXmx(nxsample_single_axis).entries[0].samples[0]
    gonio = dxtbx.nexus.get_dxtbx_goniometer(sample)
    assert isinstance(gonio, dxtbx.model.Goniometer)
    assert gonio.get_rotation_axis() == (0.0, 1.0, 0.0)
    assert gonio.get_fixed_rotation() == (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
    assert gonio.get_setting_rotation() == (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)


@pytest.fixture
def nxsample_no_depends_on():
    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        sample = f.create_group("/entry/sample")
        sample.attrs["NX_class"] = "NXsample"
        sample["depends_on"] = b"."
        yield f


def test_get_dxtbx_goniometer_sample_no_depends_on_returns_none(nxsample_no_depends_on):
    nxsample = nxmx.NXsample(nxsample_no_depends_on["/entry/sample"])
    assert dxtbx.nexus.get_dxtbx_goniometer(nxsample) is None


def test_get_dxtbx_scan_sample_no_depends_on_returns_none(
    nxsample_no_depends_on, mocker
):
    nxsample = nxmx.NXsample(nxsample_no_depends_on["/entry/sample"])
    nxdetector = mocker.Mock()
    assert dxtbx.nexus.get_dxtbx_scan(nxsample, nxdetector) is None


@pytest.fixture
def nxsample_gridscan():
    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        entry = f.create_group("/entry")
        entry.attrs["NX_class"] = "NXentry"
        entry["definition"] = "NXmx"

        sample = entry.create_group("sample")
        sample.attrs["NX_class"] = "NXsample"
        sample["name"] = "mysample"
        sample["depends_on"] = b"/entry/sample/transformations/phi"

        transformations = sample.create_group("transformations")
        transformations.attrs["NX_class"] = "NXtransformations"
        omega = transformations.create_dataset("omega", data=np.full(15, 60))
        omega.attrs["depends_on"] = b"."
        omega.attrs["transformation_type"] = b"rotation"
        omega.attrs["units"] = b"deg"
        omega.attrs["vector"] = np.array([-1.0, 0.0, 0.0])
        omega.attrs["omega_offset"] = np.array([0.0, 0.0, 0.0])

        phi = transformations.create_dataset("phi", data=np.array([0.0]))
        phi.attrs["depends_on"] = b"/entry/sample/transformations/chi"
        phi.attrs["transformation_type"] = b"rotation"
        phi.attrs["units"] = b"deg"
        phi.attrs["vector"] = np.array([-1.0, 0, 0])

        chi = transformations.create_dataset("chi", data=np.array([0.0]))
        chi.attrs["depends_on"] = b"/entry/sample/transformations/sam_x"
        chi.attrs["transformation_type"] = b"rotation"
        chi.attrs["units"] = b"deg"
        chi.attrs["vector"] = np.array([0, 0, 1])

        sam_x = transformations.create_dataset("sam_x", data=np.full(15, 300))
        sam_x.attrs["depends_on"] = b"/entry/sample/transformations/sam_y"
        sam_x.attrs["equipment_component"] = b"goniometer"
        sam_x.attrs["transformation_type"] = b"translation"
        sam_x.attrs["units"] = b"mm"
        sam_x.attrs["vector"] = np.array([1.0, 0.0, 0.0])

        sam_y = transformations.create_dataset("sam_y", data=np.arange(0, 150, 10))
        sam_y.attrs["depends_on"] = b"/entry/sample/transformations/sam_z"
        sam_y.attrs["equipment_component"] = b"goniometer"
        sam_y.attrs["transformation_type"] = b"translation"
        sam_y.attrs["units"] = b"mm"
        sam_y.attrs["vector"] = np.array([0.0, 1.0, 0.0])

        sam_z = transformations.create_dataset("sam_z", data=0)
        sam_z.attrs["depends_on"] = b"/entry/sample/transformations/omega"
        sam_z.attrs["equipment_component"] = b"goniometer"
        sam_z.attrs["transformation_type"] = b"translation"
        sam_z.attrs["units"] = b"mm"
        sam_z.attrs["vector"] = np.array([0.0, 0.0, 1.0])

        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"

        detector = instrument.create_group("detector")
        detector.attrs["NX_class"] = "NXdetector"

        yield f


def test_get_dxtbx_goniometer_grid_scan(nxsample_gridscan):
    sample = nxmx.NXmx(nxsample_gridscan).entries[0].samples[0]
    gonio = dxtbx.nexus.get_dxtbx_goniometer(sample)
    assert isinstance(gonio, dxtbx.model.MultiAxisGoniometer)
    assert gonio.get_rotation_axis() == (1.0, 0.0, 0.0)
    assert list(gonio.get_angles()) == [0.0, 0.0, 60.0]
    assert list(gonio.get_names()) == ["phi", "chi", "omega"]
    assert gonio.get_scan_axis() == 0


def test_get_dxtbx_beam(nxmx_example):
    instrument = nxmx.NXmx(nxmx_example).entries[0].instruments[0]
    beam = dxtbx.nexus.CachedWavelengthBeamFactory(instrument.beams[0]).make_beam()
    assert isinstance(beam, dxtbx.model.Beam)
    assert beam.get_wavelength() == 0.976223
    assert beam.get_sample_to_source_direction() == (0.0, 0.0, 1.0)


def test_get_dxtbx_beam_array_length_1():
    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = np.array([0.987])
        beam["incident_wavelength"].attrs["units"] = b"angstrom"

        nxbeam = nxmx.NXbeam(f["/entry/instrument/beam"])
        beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
        assert beam_factory.make_beam(index=0).get_wavelength() == 0.987


def test_get_dxtbx_beam_array():
    wavelengths = [0.9871, 0.9872, 0.9870]
    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = np.array(wavelengths)
        beam["incident_wavelength"].attrs["units"] = b"angstrom"

        nxbeam = nxmx.NXbeam(f["/entry/instrument/beam"])
        beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
        for i, w in enumerate(wavelengths):
            assert beam_factory.make_beam(index=i).get_wavelength() == w


def test_get_dxtbx_spectrum():
    # gaussian bandpass with 20eV std dev centered on 9500 eV
    energy = 9500
    channels = np.linspace(9400, 9600, 1000)
    weights = stats.norm.pdf(channels, energy, 20)

    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = factor_ev_angstrom / channels
        beam["incident_wavelength"].attrs["units"] = b"angstrom"
        beam["incident_wavelength_weights"] = weights

        nxbeam = nxmx.NXbeam(f["/entry/instrument/beam"])
        beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
        assert (
            pytest.approx(beam_factory.make_beam().get_wavelength())
            == factor_ev_angstrom / energy
        )
        assert (
            pytest.approx(beam_factory.make_spectrum().get_weighted_energy_eV())
            == energy
        )


def test_get_dxtbx_spectrum_with_variants():
    # drifting gaussian bandpass with 20eV std dev
    energies = np.array([9500, 9520, 9510])
    channels = np.linspace(9400, 9600, 1000)
    weights = np.vstack([stats.norm.pdf(channels, e, 20) for e in energies])
    # calibrated energies not necessarily matching the weighted mean of the spectra
    calibrated_energies = np.array([e + 5 for e in energies])

    with h5py.File(" ", mode="w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = factor_ev_angstrom / calibrated_energies
        beam["incident_wavelength"].attrs["units"] = b"angstrom"
        beam["incident_wavelength"].attrs["variant"] = b"incident_wavelength_1Dspectrum"
        beam["incident_wavelength_1Dspectrum"] = factor_ev_angstrom / channels
        beam["incident_wavelength_1Dspectrum"].attrs["units"] = b"angstrom"
        beam["incident_wavelength_1Dspectrum_weights"] = weights

        nxbeam = nxmx.NXbeam(f["/entry/instrument/beam"])
        beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
        for i, (calibrated_energy, energy) in enumerate(
            zip(calibrated_energies, energies)
        ):
            assert (
                pytest.approx(beam_factory.make_beam(i).get_wavelength())
                == factor_ev_angstrom / calibrated_energy
            )
            assert (
                pytest.approx(beam_factory.make_spectrum(i).get_weighted_energy_eV())
                == energy
            )


def test_get_dxtbx_scan(nxmx_example):
    sample = nxmx.NXmx(nxmx_example).entries[0].samples[0]
    instrument = nxmx.NXmx(nxmx_example).entries[0].instruments[0]
    scan = dxtbx.nexus.get_dxtbx_scan(sample, instrument.detectors[0])
    assert scan.get_num_images() == 10
    assert scan.get_image_range() == (1, 10)
    assert pytest.approx(scan.get_oscillation()) == (0.0, 0.1)
    assert pytest.approx(scan.get_oscillation_range()) == (0.0, 1.0)
    assert list(scan.get_exposure_times()) == [0.1] * 10


def test_get_dxtbx_scan_grid_scan(nxsample_gridscan):
    sample = nxmx.NXmx(nxsample_gridscan).entries[0].samples[0]
    instrument = nxmx.NXmx(nxsample_gridscan).entries[0].instruments[0]
    scan = dxtbx.nexus.get_dxtbx_scan(sample, instrument.detectors[0])
    assert scan.get_num_images() == 15
    assert scan.get_image_range() == (1, 15)
    assert scan.get_oscillation() == (0.0, 0.0)
    assert scan.get_oscillation_range() == (0.0, 0.0)
    assert list(scan.get_exposure_times()) == [0.0] * 15


def test_get_dxtbx_detector(nxmx_example):
    instrument = nxmx.NXmx(nxmx_example).entries[0].instruments[0]
    wavelength = instrument.beams[0].incident_wavelength.to("angstrom").magnitude
    detector = dxtbx.nexus.get_dxtbx_detector(instrument.detectors[0], wavelength)

    assert isinstance(detector, dxtbx.model.Detector)
    assert len(detector) == 1
    panel = detector[0]
    assert panel.get_distance() == 289.3
    assert panel.get_origin() == (-155.985, 166.904, -289.3)
    assert panel.get_material() == "Si"
    assert panel.get_pixel_size() == (0.075, 0.075)
    assert panel.get_slow_axis() == (0.0, -1.0, 0.0)
    assert panel.get_fast_axis() == (1.0, 0.0, 0.0)
    assert panel.get_image_size() == (4148, 4362)
    assert panel.get_image_size_mm() == pytest.approx((311.09999999999997, 327.15))
    assert panel.get_name() == "/entry/instrument/detector/module"
    assert panel.get_normal() == (0.0, 0.0, -1.0)
    assert panel.get_trusted_range() == (0, 9266)
    assert panel.get_type() == "SENSOR_PAD"
    px_mm = panel.get_px_mm_strategy()
    assert px_mm.t0() == panel.get_thickness() == 0.45
    assert px_mm.mu() == panel.get_mu() == pytest.approx(3.9217189904637366)


def test_get_dxtbx_detector_beam_center_fallback(nxmx_example):
    nxmx_example["/entry/instrument/detector/module/module_offset"].attrs[
        "offset"
    ] = np.array((0, 0, 0))
    instrument = nxmx.NXmx(nxmx_example).entries[0].instruments[0]
    wavelength = instrument.beams[0].incident_wavelength.to("angstrom").magnitude
    detector = dxtbx.nexus.get_dxtbx_detector(instrument.detectors[0], wavelength)
    assert detector[0].get_origin() == pytest.approx(
        (-155.985, 166.904, -289.3), rel=1e-5
    )


@pytest.fixture
def detector_with_multiple_modules():

    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:

        detector = f.create_group("/entry/instrument/detector")
        detector.attrs["NX_class"] = "NXdetector"
        detector["beam_center_x"] = 2079.79727597266
        detector["beam_center_y"] = 2225.38773853771
        detector["count_time"] = 0.00285260857097799
        detector["depends_on"] = "/entry/instrument/detector/transformations/det_z"
        detector["description"] = "Eiger 16M"
        detector["distance"] = 0.237015940260233
        detector.create_dataset("data", data=np.zeros((100, 100)))
        detector["sensor_material"] = "Silicon"
        detector["sensor_thickness"] = 0.00045
        detector["sensor_thickness"].attrs["units"] = b"m"
        detector["x_pixel_size"] = 7.5e-05
        detector["y_pixel_size"] = 7.5e-05
        detector["underload_value"] = 0
        detector["saturation_value"] = 9266
        detector["frame_time"] = 0.1
        detector["frame_time"].attrs["units"] = "s"
        detector["bit_depth_readout"] = np.array(32)
        mask = np.zeros((2, 100, 200), dtype="i8")
        detector.create_dataset("pixel_mask", data=mask)

        detector_transformations = detector.create_group("transformations")
        detector_transformations.attrs["NX_class"] = "NXtransformations"
        det_z = detector_transformations.create_dataset("det_z", data=np.array([289.3]))
        det_z.attrs["depends_on"] = b"."
        det_z.attrs["transformation_type"] = b"translation"
        det_z.attrs["units"] = b"mm"
        det_z.attrs["vector"] = np.array([0.0, 0.0, 1.0])

        def make_module(name, depends_on, data_origin, fast_direction, slow_direction):
            module = detector.create_group(name)
            module.attrs["NX_class"] = "NXdetector_module"
            module.create_dataset("data_size", data=np.array([1, 100, 200]))
            module.create_dataset("data_origin", data=np.array(data_origin))
            fast = module.create_dataset("fast_pixel_direction", data=0.075)
            fast.attrs["transformation_type"] = "translation"
            fast.attrs["depends_on"] = depends_on
            fast.attrs["vector"] = np.array(fast_direction)
            fast.attrs["units"] = "mm"
            slow = module.create_dataset("slow_pixel_direction", data=0.075)
            slow.attrs["transformation_type"] = "translation"
            slow.attrs["depends_on"] = depends_on
            slow.attrs["vector"] = np.array(slow_direction)
            slow.attrs["units"] = "mm"

        make_module(
            name="m0",
            depends_on="/entry/instrument/detector/transformations/det_z",
            data_origin=[0, 0, 0],
            fast_direction=[-0.999998, -0.001781, 0],
            slow_direction=[-0.001781, 0.999998, 0],
        )
        make_module(
            name="m1",
            depends_on="/entry/instrument/detector/transformations/det_z",
            data_origin=[1, 0, 0],
            fast_direction=[-0.999998, -0.001781, 0],
            slow_direction=[-0.001781, 0.999998, 0],
        )

        nxdata = f.create_group("/entry/data")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata.create_dataset(
            "data",
            data=np.array(
                [np.full((2, 100, 200), i, dtype=np.int32) for i in range(3)]
            ),
        )
        nxdata.attrs["signal"] = "/entry/data/data"

        yield f


def test_get_dxtbx_detector_with_multiple_modules(detector_with_multiple_modules):
    det = nxmx.NXdetector(detector_with_multiple_modules["/entry/instrument/detector"])
    wavelength = 1

    detector = dxtbx.nexus.get_dxtbx_detector(det, wavelength)
    assert len(detector) == 2
    for panel in detector:
        assert panel.get_image_size() == (200, 100)

    nxdata = nxmx.NXdata(detector_with_multiple_modules["/entry/data"])
    for i in range(3):
        raw_data = dxtbx.nexus.get_raw_data(nxdata, det, i)
        assert len(raw_data) == 2
        for module_data in raw_data:
            assert module_data.all() == (100, 200)
            assert module_data.all_eq(i)

    mask = dxtbx.nexus.get_static_mask(det)
    assert len(mask) == 2
    for module_mask in mask:
        assert isinstance(module_mask, flex.bool)
        assert module_mask.all() == (100, 200)


@pytest.fixture
def detector_with_two_theta():
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = 0.495937
        beam["incident_wavelength"].attrs["units"] = b"angstrom"

        detector = f.create_group("/entry/instrument/detector")
        detector.attrs["NX_class"] = "NXdetector"
        detector["sensor_material"] = "Silicon"
        detector["sensor_thickness"] = 0.00045
        detector["sensor_thickness"].attrs["units"] = b"m"

        module = detector.create_group("module")
        module.attrs["NX_class"] = "NXdetector_module"
        module.create_dataset("data_size", data=np.array([2162, 2068]))

        fast_pixel_direction = module.create_dataset(
            "fast_pixel_direction", data=7.5e-5
        )
        fast_pixel_direction.attrs["transformation_type"] = "translation"
        fast_pixel_direction.attrs[
            "depends_on"
        ] = "/entry/instrument/detector/module/module_offset"
        fast_pixel_direction.attrs["vector"] = np.array([-1.0, 0.0, 0.0])
        fast_pixel_direction.attrs["units"] = "m"

        slow_pixel_direction = module.create_dataset(
            "slow_pixel_direction", data=7.5e-5
        )
        slow_pixel_direction.attrs["transformation_type"] = "translation"
        slow_pixel_direction.attrs[
            "depends_on"
        ] = "/entry/instrument/detector/module/module_offset"
        slow_pixel_direction.attrs["vector"] = np.array([0, -1.0, 0.0])
        slow_pixel_direction.attrs["units"] = "m"

        module_offset = module.create_dataset("module_offset", data=112.19250476301882)
        module_offset.attrs["transformation_type"] = "translation"
        module_offset.attrs[
            "depends_on"
        ] = "/entry/instrument/detector/transformations/det_z"
        module_offset.attrs["vector"] = np.array([0.72264186, 0.69122265, 0.0])
        module_offset.attrs["units"] = b"mm"

        transformations = detector.create_group("transformations")
        det_z = transformations.create_dataset("det_z", data=120.0)
        det_z.attrs[
            "depends_on"
        ] = b"/entry/instrument/detector/transformations/two_theta"
        det_z.attrs["transformation_type"] = b"translation"
        det_z.attrs["units"] = b"mm"
        det_z.attrs["vector"] = np.array([0.0, 0.0, 1.0])

        two_theta = transformations.create_dataset("two_theta", data=45)
        two_theta.attrs["depends_on"] = b"."
        two_theta.attrs["transformation_type"] = b"rotation"
        two_theta.attrs["units"] = b"deg"
        two_theta.attrs["vector"] = np.array([-1.0, 0.0, 0.0])

        yield f


def test_get_dxtbx_detector_with_two_theta(detector_with_two_theta):
    det = nxmx.NXdetector(detector_with_two_theta["/entry/instrument/detector"])
    wavelength = detector_with_two_theta["/entry/instrument/beam/incident_wavelength"][
        ()
    ]

    detector = dxtbx.nexus.get_dxtbx_detector(det, wavelength)
    panel = detector[0]
    assert panel.get_fast_axis() == (1.0, 0.0, 0.0)
    assert panel.get_slow_axis() == (0.0, -0.7071067811865475, -0.7071067811865476)
    assert panel.get_origin() == (
        -81.07500032000678,
        139.68894494331983,
        -30.01668254145155,
    )
    assert panel.get_distance() == pytest.approx(120)


@pytest.fixture(
    params=[True, False], ids=["equipment_component", "no equipment_component"]
)
def hierarchical_detector(request):
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        beam = f.create_group("/entry/instrument/beam")
        beam.attrs["NX_class"] = "NXbeam"
        beam["incident_wavelength"] = 0.495937
        beam["incident_wavelength"].attrs["units"] = b"angstrom"

        detector = f.create_group("/entry/instrument/ELE_D0")
        detector.attrs["NX_class"] = "NXdetector"
        detector["sensor_material"] = "Si"
        detector["sensor_thickness"] = 320
        detector["sensor_thickness"].attrs["units"] = b"microns"

        def make_module(name, depends_on, data_origin, fast_direction, slow_direction):
            module = detector.create_group(name)
            module.attrs["NX_class"] = "NXdetector_module"
            module.create_dataset("data_size", data=np.array([254, 254]))
            module.create_dataset("data_origin", data=np.array(data_origin))
            fast = module.create_dataset("fast_pixel_direction", data=0.075)
            fast.attrs["transformation_type"] = "translation"
            fast.attrs["depends_on"] = depends_on
            fast.attrs["vector"] = np.array(fast_direction)
            fast.attrs["units"] = "mm"
            slow = module.create_dataset("slow_pixel_direction", data=0.075)
            slow.attrs["transformation_type"] = "translation"
            slow.attrs["depends_on"] = depends_on
            slow.attrs["vector"] = np.array(slow_direction)
            slow.attrs["units"] = "mm"

        make_module(
            name="ARRAY_D0Q0M0A0",
            depends_on="/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M0A0",
            data_origin=[1, 1],
            fast_direction=[-0.999998, -0.001781, 0],
            slow_direction=[-0.001781, 0.999998, 0],
        )
        make_module(
            name="ARRAY_D0Q0M0A1",
            depends_on="/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M0A1",
            data_origin=[1, 259],
            fast_direction=[-0.999998, -0.001781, 0],
            slow_direction=[-0.001781, 0.999998, 0],
        )
        make_module(
            name="ARRAY_D0Q0M12A0",
            depends_on="/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M12A0",
            data_origin=[6169, 1],
            fast_direction=[-0.999996, 0.00279501, 0],
            slow_direction=[0.00279501, 0.999996, 0],
        )
        make_module(
            name="ARRAY_D0Q0M12A1",
            depends_on="/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M12A1",
            data_origin=[6169, 259],
            fast_direction=[-0.999996, 0.00279501, 0],
            slow_direction=[0.00279501, 0.999996, 0],
        )

        transformations = detector.create_group("transformations")

        t = transformations.create_dataset("AXIS_D0Q0M0A0", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0M0"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([38.5842, -19.1314, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0M0A1", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0M0"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([19.2342, -19.1658, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0M12A0", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0M12"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([38.4963, -19.3077, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0M12A1", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0M12"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([19.1463, -19.2536, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0M0", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([38.3107, -61.1138, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0M12", data=0)
        t.attrs["depends_on"] = b"AXIS_D0Q0"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([39.1843, 61.4038, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0Q0", data=0)
        t.attrs["depends_on"] = b"AXIS_D0"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([75.1333, -84.6597, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"

        t = transformations.create_dataset("AXIS_D0", data=0)
        t.attrs["depends_on"] = b"AXIS_RAIL"
        t.attrs["transformation_type"] = b"rotation"
        t.attrs["offset"] = np.array([-0.429033, 0.121021, -0.294])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, -1])
        t.attrs["units"] = b"degrees"
        if request.param:
            t.attrs["equipment_component"] = "detector_arm"

        t = transformations.create_dataset("AXIS_RAIL", data=97.83)
        t.attrs["depends_on"] = b"."
        t.attrs["transformation_type"] = b"translation"
        t.attrs["offset"] = np.array([0, 0, 0])
        t.attrs["offset_units"] = b"mm"
        t.attrs["vector"] = np.array([0, 0, 1])
        t.attrs["units"] = b"mm"
        if request.param:
            t.attrs["equipment_component"] = "detector_arm"

        yield f


def test_get_dxtbx_detector_hierarchical(hierarchical_detector):
    det = nxmx.NXdetector(hierarchical_detector["/entry/instrument/ELE_D0"])
    wavelength = hierarchical_detector["/entry/instrument/beam/incident_wavelength"][()]

    detector = dxtbx.nexus.get_dxtbx_detector(det, wavelength)
    assert len(detector) == 4

    assert detector[0].get_name() == "/entry/instrument/ELE_D0/ARRAY_D0Q0M0A0"
    assert detector[1].get_name() == "/entry/instrument/ELE_D0/ARRAY_D0Q0M0A1"
    assert detector[2].get_name() == "/entry/instrument/ELE_D0/ARRAY_D0Q0M12A0"
    assert detector[3].get_name() == "/entry/instrument/ELE_D0/ARRAY_D0Q0M12A1"

    assert (
        detector[0].parent().parent().get_name()
        == "/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M0"
    )
    assert (
        detector[1].parent().parent().get_name()
        == "/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M0"
    )
    assert (
        detector[2].parent().parent().get_name()
        == "/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M12"
    )
    assert (
        detector[3].parent().parent().get_name()
        == "/entry/instrument/ELE_D0/transformations/AXIS_D0Q0M12"
    )
    for i in range(4):
        assert (
            detector[i].parent().parent().parent().get_name()
            == "/entry/instrument/ELE_D0/transformations/AXIS_D0Q0"
        )
        assert (
            detector[i].parent().parent().parent().parent().get_name()
            == "/entry/instrument/ELE_D0/transformations/AXIS_D0"
        )

    assert detector[0].get_origin() == (-151.599167, -164.783879, -97.536)
    assert detector[0].get_local_origin() == (0.0, 0.0, 0.0)
    fast_axis = pytest.approx((0.9999984140169291, -0.0017810007373656254, 0.0))
    assert detector[0].get_fast_axis() == fast_axis
    assert detector[0].get_local_fast_axis() == fast_axis
    slow_axis = pytest.approx((0.0017810007373656254, 0.9999984140169291, 0.0))
    assert detector[0].get_slow_axis() == slow_axis
    assert detector[0].get_local_slow_axis() == slow_axis
    assert detector[0].get_distance() == pytest.approx(97.536)

    depth = 0
    pg = detector.hierarchy()
    while True:
        depth += 1
        if pg.is_panel():
            break
        else:
            pg = pg[0]
    if (
        "equipment_component"
        in hierarchical_detector[
            "/entry/instrument/ELE_D0/transformations/AXIS_RAIL"
        ].attrs
    ):
        assert depth == 5
    else:
        assert depth == 6


@pytest.fixture
def pixel_mask_example():
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        detector = f.create_group("/entry/instrument/detector")
        detector.attrs["NX_class"] = "NXdetector"

        module = detector.create_group("module")
        module.attrs["NX_class"] = "NXdetector_module"
        module.create_dataset("data_origin", data=np.array([0.0, 0.0]))
        module.create_dataset("data_size", data=np.array([4362, 4148]))

        nf = 1028  # module pixels fast
        ns = 512  # module pixels slow
        df = 12  # module gap fast
        ds = 38  # module gap slow

        mask = np.zeros((4362, 4148), dtype="i8")
        for j in range(mask.shape[1] // (nf + df)):
            mask[:, (j + 1) * nf + j * df : (j + 1) * (nf + df)] = 1
        for i in range(mask.shape[0] // (ns + ds)):
            mask[(i + 1) * ns + i * ds : (i + 1) * (ns + ds), :] = 1

        detector.create_dataset("pixel_mask", data=mask)
        yield f


def test_get_static_mask(pixel_mask_example):
    det = nxmx.NXdetector(pixel_mask_example["/entry/instrument/detector"])
    mask = dxtbx.nexus.get_static_mask(det)
    assert len(mask) == 1
    assert isinstance(mask[0], flex.bool)
    assert mask[0].all() == (4362, 4148)
    assert np.all(
        mask[0].as_numpy_array()
        == (pixel_mask_example["/entry/instrument/detector/pixel_mask"][()] == 0)
    )


@pytest.fixture
def nxdata_example():
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        detector = f.create_group("/entry/instrument/detector")
        detector.attrs["NX_class"] = "NXdetector"

        module = detector.create_group("module")
        module.attrs["NX_class"] = "NXdetector_module"
        module.create_dataset("data_origin", data=np.array([0.0, 0.0]))
        module.create_dataset("data_size", data=np.array([4362, 4148]))

        nxdata = f.create_group("/entry/data")
        nxdata.attrs["NX_class"] = "NXdata"
        nxdata.create_dataset(
            "data",
            data=np.array([np.full((4362, 4148), i, dtype=np.int32) for i in range(3)]),
        )
        nxdata.attrs["signal"] = "/entry/data/data"

        yield f


def test_get_raw_data_single_panel(nxdata_example):
    det = nxmx.NXdetector(nxdata_example["/entry/instrument/detector"])
    nxdata = nxmx.NXdata(nxdata_example["/entry/data"])
    for i in range(3):
        raw_data = dxtbx.nexus.get_raw_data(nxdata, det, i)
        assert len(raw_data) == 1
        assert isinstance(raw_data[0], flex.int)
        assert raw_data[0].all() == (4362, 4148)
        assert raw_data[0].all_eq(i)


def test_dataset_as_flex_int():
    slices = ()
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        g = f.create_group("/foo")
        g.create_dataset("int32", data=np.array([0, 1], dtype=np.int32))
        g.create_dataset("intc", data=np.array([1, 2], dtype=np.intc))
        g.create_dataset("int64", data=np.array([2, 3], dtype=np.int64))
        for k in ("int32", "intc"):
            flex_a = dxtbx.nexus._dataset_as_flex(g[k], slices)
            assert isinstance(flex_a, flex.int)
            assert flex_a.all() == g[k].shape
            assert list(flex_a) == list(g[k])
        with pytest.raises(TypeError, match="Unsupported integer dtype .*"):
            dxtbx.nexus._dataset_as_flex(g["int64"], slices)
        flex_a = dxtbx.nexus._dataset_as_flex(g["int64"], slices, bit_depth=32)
        assert isinstance(flex_a, flex.int)
        assert flex_a.all() == g["int64"].shape
        assert list(flex_a) == list(g["int64"])


def test_dataset_as_flex_float():
    slices = ()
    np_float_types = (np.half, np.single, np.float16, np.float32)
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        g = f.create_group("/foo")
        for i, dtype in enumerate(np_float_types):
            d = g.create_dataset(f"float-{i}", data=np.array([0, 1], dtype=dtype))
            flex_a = dxtbx.nexus._dataset_as_flex(d, slices)
            assert isinstance(flex_a, flex.float)
            assert flex_a.all() == d.shape
            assert list(flex_a) == list(d)


def test_dataset_as_flex_double():
    slices = ()
    np_double_types = (
        np.float_,
        np.double,
        np.float64,
    )
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        g = f.create_group("/foo")
        for i, dtype in enumerate(np_double_types):
            d = g.create_dataset(f"double-{i}", data=np.array([0, 1], dtype=dtype))
            flex_a = dxtbx.nexus._dataset_as_flex(d, slices)
            assert isinstance(flex_a, flex.double)
            assert flex_a.all() == d.shape
            assert list(flex_a) == list(d)


def test_dataset_as_flex_unsupported():
    slices = ()
    with h5py.File(" ", "w", **pytest.h5_in_memory) as f:
        g = f.create_group("/foo")
        d = g.create_dataset("bool", data=np.array([0, 1], dtype=bool))
        with pytest.raises(TypeError, match="Unsupported dtype .*"):
            dxtbx.nexus._dataset_as_flex(d, slices)



from __future__ import annotations

import os


def resolve_path(path, directory=None):
    """Resolve a file path.

    First expand any environment and user variables. Then create the absolute
    path by applying the relative path to the provided directory, if necessary.

    Args:
      path (str): The path to resolve
      directory (Optional[str]): The local path to resolve relative links

    Returns:
        str: The absolute path to the file to read
    """
    if not path:
        return ""
    path = os.path.expanduser(os.path.expandvars(path))
    if directory and not os.path.isabs(path):
        path = os.path.join(directory, path)
    return os.path.abspath(path)



from __future__ import annotations

import os
import pickle

from dxtbx.format.image import ImageBool, ImageDouble
from dxtbx.imageset import ImageSequence, ImageSet, ImageSetFactory
from dxtbx.model import BeamFactory, DetectorFactory, GoniometerFactory, ScanFactory
from dxtbx.serialize.filename import resolve_path


def filename_to_absolute(filename):
    """Convert filenames to absolute form."""

    if isinstance(filename, list):
        return [os.path.abspath(f) for f in filename]

    return os.path.abspath(filename)


def filename_or_none(filename):
    if filename is None or filename == "":
        return None
    return filename_to_absolute(filename)


def basic_imageset_to_dict(imageset):
    """Convert an imageset to a dictionary

    Params:
        imageset The imageset

    Returns:
        A dictionary of the parameters

    """

    return {
        "__id__": "imageset",
        "filenames": filename_to_absolute(imageset.paths()),
        "mask": filename_or_none(imageset.external_lookup.mask.filename),
        "gain": filename_or_none(imageset.external_lookup.gain.filename),
        "pedestal": filename_or_none(imageset.external_lookup.pedestal.filename),
        "beam": imageset.get_beam(0).to_dict(),
        "detector": imageset.get_detector(0).to_dict(),
    }


def imagesequence_to_dict(sequence):
    """Convert a sequence to a dictionary

    Params:
        sequence The sequence

    Returns:
        A dictionary of the parameters

    """

    return {
        "__id__": "imageset",
        "template": filename_to_absolute(sequence.get_template()),
        "mask": filename_or_none(sequence.external_lookup.mask.filename),
        "gain": filename_or_none(sequence.external_lookup.gain.filename),
        "pedestal": filename_or_none(sequence.external_lookup.pedestal.filename),
        "beam": sequence.get_beam().to_dict(),
        "detector": sequence.get_detector().to_dict(),
        "goniometer": sequence.get_goniometer().to_dict(),
        "scan": sequence.get_scan().to_dict(),
    }


def imageset_to_dict(imageset):
    """Convert the imageset to a dictionary

    Params:
        imageset The imageset

    Returns:
        A dictionary of the parameters

    """
    # If this is an imageset then return a list of filenames
    if isinstance(imageset, ImageSequence):
        return imagesequence_to_dict(imageset)
    elif isinstance(imageset, ImageSet):
        return basic_imageset_to_dict(imageset)
    else:
        raise TypeError("Unknown ImageSet Type")


def basic_imageset_from_dict(d, directory=None):
    """Construct an ImageSet class from the dictionary."""
    # Get the filename list and create the imageset
    filenames = [resolve_path(str(p), directory=directory) for p in d["filenames"]]
    imageset = ImageSetFactory.new(filenames)[0]

    # Set some external lookups
    if "mask" in d and d["mask"] is not None and d["mask"] != "":
        path = resolve_path(d["mask"], directory=directory)
        with open(path) as infile:
            imageset.external_lookup.mask.filename = path
            imageset.external_lookup.mask.data = ImageBool(pickle.load(infile))
    if "gain" in d and d["gain"] is not None and d["gain"] != "":
        path = resolve_path(d["gain"], directory=directory)
        with open(path) as infile:
            imageset.external_lookup.gain.filename = path
            imageset.external_lookup.gain.data = ImageDouble(pickle.load(infile))
    if "pedestal" in d and d["pedestal"] is not None and d["pedestal"] != "":
        path = resolve_path(d["pedestal"], directory=directory)
        with open(path) as infile:
            imageset.external_lookup.pedestal.filename = path
            imageset.external_lookup.pedestal.data = ImageDouble(pickle.load(infile))

    # Get the existing models as dictionaries
    beam_dict = imageset.get_beam(0).to_dict()
    detector_dict = imageset.get_detector(0).to_dict()

    # Set models
    imageset.set_beam(BeamFactory.from_dict(d.get("beam"), beam_dict))
    imageset.set_detector(DetectorFactory.from_dict(d.get("detector"), detector_dict))

    return imageset


def imagesequence_from_dict(d, check_format=True, directory=None):
    """Construct and image sequence from the dictionary."""
    # Get the template (required)
    template = resolve_path(str(d["template"]), directory=directory)

    # If the scan isn't set, find all available files
    scan_dict = d.get("scan")
    if scan_dict is None:
        image_range = None
    else:
        image_range = scan_dict.get("image_range")

    # Set the models with the existing models as templates
    beam = BeamFactory.from_dict(d.get("beam"))
    goniometer = GoniometerFactory.from_dict(d.get("goniometer"))
    detector = DetectorFactory.from_dict(d.get("detector"))
    scan = ScanFactory.from_dict(d.get("scan"))

    # Construct the sequence
    try:
        sequence = ImageSetFactory.from_template(
            template,
            image_range,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            check_format=check_format,
        )[0]
    except Exception:
        indices = list(range(image_range[0], image_range[1] + 1))
        sequence = ImageSetFactory.make_sequence(
            template,
            indices,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            check_format=check_format,
        )

    # Set some external lookups
    if "mask" in d and d["mask"] is not None and d["mask"] != "":
        path = resolve_path(d["mask"], directory=directory)
        with open(path) as infile:
            sequence.external_lookup.mask.filename = path
            sequence.external_lookup.mask.data = ImageBool(pickle.load(infile))
    if "gain" in d and d["gain"] is not None and d["gain"] != "":
        path = resolve_path(d["gain"], directory=directory)
        with open(path) as infile:
            sequence.external_lookup.gain.filename = path
            sequence.external_lookup.gain.data = ImageDouble(pickle.load(infile))
    if "pedestal" in d and d["pedestal"] is not None and d["pedestal"] != "":
        path = resolve_path(d["pedestal"], directory=directory)
        with open(path) as infile:
            sequence.external_lookup.pedestal.filename = path
            sequence.external_lookup.pedestal.data = ImageDouble(pickle.load(infile))

    return sequence


def imageset_from_dict(d, check_format=True, directory=None):
    """Convert the dictionary to a sequence

    Params:
        d The dictionary of parameters

    Returns:
        The sequence

    """
    # Check the input
    if d is None:
        return None

    # Check the version and id
    if str(d["__id__"]) != "imageset":
        raise ValueError('"__id__" does not equal "imageset"')

    if "filenames" in d:
        return basic_imageset_from_dict(d, directory=directory)
    elif "template" in d:
        return imagesequence_from_dict(
            d, check_format=check_format, directory=directory
        )
    else:
        raise TypeError("Unable to deserialize given imageset")



from __future__ import annotations

from pathlib import Path

from dxtbx.model.beam import BeamFactory


def test_beam():
    image = str(Path(__file__).parent / "phi_scan_001.cbf")
    assert BeamFactory.imgCIF(image)



"""
An implementation of the CBF image reader for Pilatus images, from the Pilatus
6M SN 100 currently on Diamond I03.
"""


from __future__ import annotations

import binascii
import sys

from scitbx import matrix

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatusDLS6MSN126(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 126 @ DLS."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0126" in header
            ):
                return True

        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""
        super().__init__(image_file, **kwargs)

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        if "Phi" in self._cif_header_dictionary:
            phi_value = float(self._cif_header_dictionary["Phi"].split()[0])
        else:
            phi_value = 0.0

        if "Kappa" in self._cif_header_dictionary:
            kappa_value = float(self._cif_header_dictionary["Kappa"].split()[0])
        else:
            kappa_value = 0.0

        axis = matrix.col((1, 0, 0))
        phi = matrix.col((1, 0, 0))
        kappa = matrix.col((0.914, 0.279, -0.297))

        Rphi = phi.axis_and_angle_as_r3_rotation_matrix(phi_value, deg=True)
        Rkappa = kappa.axis_and_angle_as_r3_rotation_matrix(kappa_value, deg=True)
        fixed = Rkappa * Rphi

        return self._goniometer_factory.make_goniometer(axis, fixed)

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDLS6MSN126.understand(arg))



from __future__ import annotations

import copy
import json
import os
import sys

from orderedset import OrderedSet

import boost_adaptbx.boost.python
import cctbx.crystal
import cctbx.sgtbx
import cctbx.uctbx
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.imageset import ImageGrid, ImageSequence, ImageSet
from dxtbx.model.beam import BeamFactory
from dxtbx.model.crystal import CrystalFactory
from dxtbx.model.detector import DetectorFactory
from dxtbx.model.goniometer import GoniometerFactory
from dxtbx.model.profile import ProfileModelFactory
from dxtbx.model.scan import ScanFactory
from dxtbx.util import AutoEncoder, format_float_with_standard_uncertainty

try:
    from ..dxtbx_model_ext import (
        Beam,
        BeamBase,
        Crystal,
        CrystalBase,
        Detector,
        DetectorNode,
        Experiment,
        ExperimentList,
        Goniometer,
        GoniometerBase,
        KappaDirection,
        KappaGoniometer,
        KappaScanAxis,
        MosaicCrystalKabsch2010,
        MosaicCrystalSauter2014,
        MultiAxisGoniometer,
        OffsetParallaxCorrectedPxMmStrategy,
        OffsetPxMmStrategy,
        Panel,
        ParallaxCorrectedPxMmStrategy,
        PolychromaticBeam,
        PxMmStrategy,
        Scan,
        ScanBase,
        SimplePxMmStrategy,
        Spectrum,
        VirtualPanel,
        VirtualPanelFrame,
        get_mod2pi_angles_in_range,
        get_range_of_mod2pi_angles,
        is_angle_in_range,
        parallax_correction,
        parallax_correction_inv,
    )
except ModuleNotFoundError:
    from dxtbx_model_ext import (  # type: ignore
        Beam,
        BeamBase,
        Crystal,
        CrystalBase,
        Detector,
        DetectorNode,
        Experiment,
        ExperimentList,
        Goniometer,
        GoniometerBase,
        KappaDirection,
        KappaGoniometer,
        KappaScanAxis,
        MosaicCrystalKabsch2010,
        MosaicCrystalSauter2014,
        MultiAxisGoniometer,
        OffsetParallaxCorrectedPxMmStrategy,
        OffsetPxMmStrategy,
        Panel,
        ParallaxCorrectedPxMmStrategy,
        PolychromaticBeam,
        PxMmStrategy,
        Scan,
        ScanBase,
        SimplePxMmStrategy,
        Spectrum,
        VirtualPanel,
        VirtualPanelFrame,
        get_mod2pi_angles_in_range,
        get_range_of_mod2pi_angles,
        is_angle_in_range,
        parallax_correction,
        parallax_correction_inv,
    )

__all__ = (
    "Beam",
    "BeamBase",
    "PolychromaticBeam",
    "BeamFactory",
    "Crystal",
    "CrystalBase",
    "CrystalFactory",
    "Detector",
    "DetectorFactory",
    "DetectorNode",
    "Experiment",
    "ExperimentList",
    "Goniometer",
    "GoniometerBase",
    "GoniometerFactory",
    "ImageGrid",
    "ImageSet",
    "ImageSequence",
    "KappaDirection",
    "KappaGoniometer",
    "KappaScanAxis",
    "MosaicCrystalKabsch2010",
    "MosaicCrystalSauter2014",
    "MultiAxisGoniometer",
    "OffsetParallaxCorrectedPxMmStrategy",
    "OffsetPxMmStrategy",
    "Panel",
    "ParallaxCorrectedPxMmStrategy",
    "ProfileModelFactory",
    "PxMmStrategy",
    "Scan",
    "ScanBase",
    "ScanFactory",
    "SimplePxMmStrategy",
    "Spectrum",
    "VirtualPanel",
    "VirtualPanelFrame",
    "get_mod2pi_angles_in_range",
    "get_range_of_mod2pi_angles",
    "is_angle_in_range",
    "parallax_correction",
    "parallax_correction_inv",
)


@boost_adaptbx.boost.python.inject_into(Detector)
class _detector:
    def iter_panels(self):
        """Iterate through just the panels depth-first."""
        for obj in self.iter_preorder():
            if obj.is_panel():
                yield obj

    def iter_preorder(self):
        """Iterate through the groups and panels depth-first."""
        stack = [self.hierarchy()]
        while stack:
            node = stack.pop()
            yield node
            if node.is_group():
                stack.extend(reversed(node))

    def iter_levelorder(self):
        """Iterate through the groups and panels breadth-first."""
        queue = [self.hierarchy()]
        while queue:
            node = queue.pop(0)
            yield node
            if node.is_group():
                queue.extend(node)


@boost_adaptbx.boost.python.inject_into(Crystal)
class _crystal:
    def show(self, show_scan_varying=False, out=None):
        if out is None:
            out = sys.stdout
        print(self.as_str(show_scan_varying=show_scan_varying), file=out)

    def get_crystal_symmetry(self, assert_is_compatible_unit_cell=True):
        return cctbx.crystal.symmetry(
            unit_cell=self.get_unit_cell(),
            space_group=self.get_space_group(),
            assert_is_compatible_unit_cell=assert_is_compatible_unit_cell,
        )

    def as_str(self, show_scan_varying=False):
        uc = self.get_unit_cell().parameters()
        uc_sd = self.get_cell_parameter_sd()
        sg = str(self.get_space_group().info())
        umat = (
            matrix.sqr(self.get_U())
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )
        bmat = (
            matrix.sqr(self.get_B())
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )
        amat = (
            (matrix.sqr(self.get_U()) * matrix.sqr(self.get_B()))
            .mathematica_form(format="% 5.4f", one_row_per_line=True)
            .splitlines()
        )

        msg = ["Crystal:"]

        if len(uc_sd) != 0:
            cell_str = [
                format_float_with_standard_uncertainty(v, e, minimum=1.0e-5)
                for (v, e) in zip(uc, uc_sd)
            ]
            msg.append("    Unit cell: " + ", ".join(cell_str))
        else:
            msg.append(
                "    Unit cell: " + "%5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f" % uc
            )
        msg.append("    Space group: " + sg)
        msg.append("    U matrix:  " + umat[0])
        msg.append("               " + umat[1])
        msg.append("               " + umat[2])
        msg.append("    B matrix:  " + bmat[0])
        msg.append("               " + bmat[1])
        msg.append("               " + bmat[2])
        msg.append("    A = UB:    " + amat[0])
        msg.append("               " + amat[1])
        msg.append("               " + amat[2])
        if self.num_scan_points > 0:
            msg.append("    A sampled at " + str(self.num_scan_points) + " scan points")
            if show_scan_varying:
                for i in range(self.num_scan_points):
                    A = matrix.sqr(self.get_A_at_scan_point(i))
                    B = matrix.sqr(self.get_B_at_scan_point(i))
                    U = matrix.sqr(self.get_U_at_scan_point(i))
                    uc = self.get_unit_cell_at_scan_point(i).parameters()
                    umat = U.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    bmat = B.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    amat = A.mathematica_form(
                        format="% 5.4f", one_row_per_line=True
                    ).splitlines()
                    msg.append("  Scan point #%i:" % (i + 1))
                    msg.append(
                        "    Unit cell: "
                        + "%5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f" % uc
                    )
                    msg.append("    U matrix:  " + umat[0])
                    msg.append("               " + umat[1])
                    msg.append("               " + umat[2])
                    msg.append("    B matrix:  " + bmat[0])
                    msg.append("               " + bmat[1])
                    msg.append("               " + bmat[2])
                    msg.append("    A = UB:    " + amat[0])
                    msg.append("               " + amat[1])
                    msg.append("               " + amat[2])

        uc = self.get_recalculated_unit_cell()
        if uc is not None:
            uc = uc.parameters()
            uc_sd = self.get_recalculated_cell_parameter_sd()
            if len(uc_sd) != 0:
                cell_str = [
                    format_float_with_standard_uncertainty(v, e, minimum=1.0e-5)
                    for (v, e) in zip(uc, uc_sd)
                ]
                msg.append("    Recalculated unit cell: " + ", ".join(cell_str))
            else:
                msg.append(
                    "    Recalculated unit cell: "
                    + "(%5.3f, %5.3f, %5.3f, %5.3f, %5.3f, %5.3f)" % uc
                )
        return "\n".join(msg)

    def __str__(self):
        return self.as_str()

    def to_dict(self):
        """Convert the crystal model to a dictionary

        Returns:
            A dictionary of the parameters
        """
        # Get the real space vectors
        A = matrix.sqr(self.get_A()).inverse()
        real_space_a = (A[0], A[1], A[2])
        real_space_b = (A[3], A[4], A[5])
        real_space_c = (A[6], A[7], A[8])

        # Get the space group Hall symbol
        hall = self.get_space_group().info().type().hall_symbol()

        # Isoforms used for stills
        try:
            identified_isoform = self.identified_isoform
        except AttributeError:
            identified_isoform = None

        # Collect the information as a python dictionary
        xl_dict = {
            "__id__": "crystal",
            "real_space_a": real_space_a,
            "real_space_b": real_space_b,
            "real_space_c": real_space_c,
            "space_group_hall_symbol": hall,
        }

        if identified_isoform is not None:
            xl_dict["identified_isoform"] = identified_isoform

        # Add in scan points if present
        if self.num_scan_points > 0:
            A_at_scan_points = tuple(
                [self.get_A_at_scan_point(i) for i in range(self.num_scan_points)]
            )
            xl_dict["A_at_scan_points"] = A_at_scan_points

        # Add in covariance of B if present
        cov_B = tuple(self.get_B_covariance())
        if len(cov_B) != 0:
            xl_dict["B_covariance"] = cov_B

        # Add in covariance of B at scan points if present
        if self.num_scan_points > 0:
            try:
                cov_B_at_scan_points = tuple(
                    [
                        tuple(self.get_B_covariance_at_scan_point(i))
                        for i in range(self.num_scan_points)
                    ]
                )
                xl_dict["B_covariance_at_scan_points"] = cov_B_at_scan_points
            except RuntimeError:
                pass

        recalculated_unit_cell = self.get_recalculated_unit_cell()
        if recalculated_unit_cell is not None:
            xl_dict["recalculated_unit_cell"] = recalculated_unit_cell.parameters()
            xl_dict[
                "recalculated_cell_parameter_sd"
            ] = self.get_recalculated_cell_parameter_sd()
            xl_dict[
                "recalculated_cell_volume_sd"
            ] = self.get_recalculated_cell_volume_sd()

        return xl_dict

    @staticmethod
    def from_dict(d):
        """Convert the dictionary to a crystal model

        Params:
            d The dictionary of parameters

        Returns:
            The crystal model
        """
        # If None, return None
        if d is None:
            return None

        # Check the version and id
        if str(d["__id__"]) != "crystal":
            raise ValueError('"__id__" does not equal "crystal"')

        # Extract from the dictionary
        real_space_a = d["real_space_a"]
        real_space_b = d["real_space_b"]
        real_space_c = d["real_space_c"]
        # str required to force unicode to ascii conversion
        space_group = str("Hall:" + d["space_group_hall_symbol"])
        xl = Crystal(
            real_space_a, real_space_b, real_space_c, space_group_symbol=space_group
        )

        # Isoforms used for stills
        try:
            xl.identified_isoform = d["identified_isoform"]
        except KeyError:
            pass

        # Extract scan point setting matrices, if present
        try:
            A_at_scan_points = d["A_at_scan_points"]
            xl.set_A_at_scan_points(A_at_scan_points)
        except KeyError:
            pass

        # Extract covariance of B, if present
        try:
            cov_B = d["B_covariance"]
            xl.set_B_covariance(cov_B)
        except KeyError:
            pass

        # Extract covariance of B at scan points, if present
        cov_B_at_scan_points = d.get("B_covariance_at_scan_points")
        if cov_B_at_scan_points is not None:
            cov_B_at_scan_points = flex.double(cov_B_at_scan_points).as_1d()
            cov_B_at_scan_points.reshape(flex.grid(xl.num_scan_points, 9, 9))
            xl.set_B_covariance_at_scan_points(cov_B_at_scan_points)

        recalculated_unit_cell = d.get("recalculated_unit_cell")
        if recalculated_unit_cell is not None:
            xl.set_recalculated_unit_cell(cctbx.uctbx.unit_cell(recalculated_unit_cell))

        recalculated_cell_parameter_sd = d.get("recalculated_cell_parameter_sd")
        if recalculated_cell_parameter_sd is not None:
            xl.set_recalculated_cell_parameter_sd(recalculated_cell_parameter_sd)

        recalculated_cell_volume_sd = d.get("recalculated_cell_volume_sd")
        if recalculated_cell_volume_sd is not None:
            xl.set_recalculated_cell_volume_sd(recalculated_cell_volume_sd)

        return xl


@boost_adaptbx.boost.python.inject_into(MosaicCrystalKabsch2010)
class _crystal_kabsch:
    def as_str(self, show_scan_varying=False):
        return "\n".join(
            (
                super(MosaicCrystalKabsch2010, self).as_str(
                    show_scan_varying=show_scan_varying
                ),
                "    Mosaicity:  %.6f" % self.get_mosaicity(),
            )
        )

    def to_dict(self):
        """Convert the crystal model to a dictionary

        Returns:
            A dictionary of the parameters

        """
        xl_dict = super(MosaicCrystalKabsch2010, self).to_dict()

        # Get the mosaicity
        mosaicity = self.get_mosaicity()
        xl_dict["mosaicity"] = mosaicity

        return xl_dict

    @classmethod
    def from_dict(cls, d):
        """Convert the dictionary to a crystal model

        Params:
            d The dictionary of parameters

        Returns:
            The crystal model

        """
        xl = MosaicCrystalKabsch2010(super(MosaicCrystalKabsch2010, cls).from_dict(d))

        # This parameter doesn't survive the Crystal copy constructor so has to be re-set.
        # Isoforms used for stills
        try:
            xl.identified_isoform = d["identified_isoform"]
        except KeyError:
            pass

        # Extract mosaicity, if present
        try:
            mosaicity = d["mosaicity"]
            xl.set_mosaicity(mosaicity)
        except KeyError:
            pass

        return xl


@boost_adaptbx.boost.python.inject_into(MosaicCrystalSauter2014)
class _crystalsauter:
    def as_str(self, show_scan_varying=False):
        return "\n".join(
            (
                super(MosaicCrystalSauter2014, self).as_str(
                    show_scan_varying=show_scan_varying
                ),
                "    Half mosaic angle (degrees):  %.6f"
                % self.get_half_mosaicity_deg(),
                "    Domain size (Angstroms):  %.6f" % self.get_domain_size_ang(),
            )
        )

    def get_A_as_sqr(self):  # required for lunus
        return matrix.sqr(self.get_A())

    def get_A_inverse_as_sqr(self):
        return self.get_A_as_sqr().inverse()

    def to_dict(self):
        """Convert the crystal model to a dictionary

        Returns:
            A dictionary of the parameters

        """
        xl_dict = super(MosaicCrystalSauter2014, self).to_dict()

        # Get the mosaic parameters
        half_mosaicity = self.get_half_mosaicity_deg()
        xl_dict["ML_half_mosaicity_deg"] = half_mosaicity

        domain_size = self.get_domain_size_ang()
        xl_dict["ML_domain_size_ang"] = domain_size

        return xl_dict

    @classmethod
    def from_dict(cls, d):
        """Convert the dictionary to a crystal model

        Params:
            d The dictionary of parameters

        Returns:
            The crystal model

        """
        xl = MosaicCrystalSauter2014(super(MosaicCrystalSauter2014, cls).from_dict(d))

        # Parameters for maximum likelihood values
        try:
            xl.set_half_mosaicity_deg(d["ML_half_mosaicity_deg"])
        except KeyError:
            pass
        try:
            xl.set_domain_size_ang(d["ML_domain_size_ang"])
        except KeyError:
            pass

        # This parameter doesn't survive the Crystal copy constructor so have to be re-set
        # Isoforms used for stills
        try:
            xl.identified_isoform = d["identified_isoform"]
        except KeyError:
            pass

        return xl


@boost_adaptbx.boost.python.inject_into(Experiment)
class _experiment:
    def load_models(self, index=None):
        """Load the models from the imageset"""
        if index is None:
            index = 0
        self.beam = self.imageset.get_beam(index)
        self.detector = self.imageset.get_detector(index)
        self.goniometer = self.imageset.get_goniometer(index)
        self.scan = self.imageset.get_scan(index)


@boost_adaptbx.boost.python.inject_into(ExperimentList)
class _experimentlist:
    def __repr__(self):
        if len(self):
            return "ExperimentList([{}])".format(", ".join(repr(x) for x in self))
        else:
            return "ExperimentList()"

    def beams(self):
        """Get a list of the unique beams (includes None)."""
        return list(OrderedSet(e.beam for e in self))

    def detectors(self):
        """Get a list of the unique detectors (includes None)."""
        return list(OrderedSet(e.detector for e in self))

    def goniometers(self):
        """Get a list of the unique goniometers (includes None)."""
        return list(OrderedSet(e.goniometer for e in self))

    def scans(self):
        """Get a list of the unique scans (includes None)."""
        return list(OrderedSet(e.scan for e in self))

    def crystals(self):
        """Get a list of the unique crystals (includes None)."""
        return list(OrderedSet(e.crystal for e in self))

    def profiles(self):
        """Get a list of the unique profile models (includes None)."""
        return list(OrderedSet(e.profile for e in self))

    def scaling_models(self):
        """Get a list of the unique scaling models (includes None)."""
        return list(OrderedSet(e.scaling_model for e in self))

    def imagesets(self):
        """Get a list of the unique imagesets."""
        return list(OrderedSet([e.imageset for e in self if e.imageset is not None]))

    def all_stills(self):
        """Check if all the experiments are stills"""
        return all(exp.is_still() for exp in self)

    def all_sequences(self):
        """Check if all the experiments are from sequences"""
        return all(exp.is_sequence() for exp in self)

    def to_dict(self):
        """Serialize the experiment list to dictionary."""

        def abspath_or_none(filename):
            if filename is None or filename == "":
                return None
            return os.path.abspath(filename)

        # Check the experiment list is consistent
        assert self.is_consistent()

        # Table of names, source collections and experiment accessors
        # for member models that are serialized as an index table-lookup
        lookup_members = [
            ("beam", self.beams, lambda x: x.beam),
            ("detector", self.detectors, lambda x: x.detector),
            ("goniometer", self.goniometers, lambda x: x.goniometer),
            ("scan", self.scans, lambda x: x.scan),
            ("crystal", self.crystals, lambda x: x.crystal),
            ("profile", self.profiles, lambda x: x.profile),
            ("scaling_model", self.scaling_models, lambda x: x.scaling_model),
            ("imageset", self.imagesets, lambda x: x.imageset),
        ]
        # Generate index lookup tables for each output member collection instance
        index_lookup = {
            name: {
                model: i for i, model in enumerate(x for x in models() if x is not None)
            }
            for name, models, _ in lookup_members
        }

        # Create the output dictionary
        result = {
            "__id__": "ExperimentList",
            "experiment": [],
        }

        # Add the experiments to the dictionary
        for e in self:
            obj = {
                "__id__": "Experiment",
                "identifier": e.identifier,
            }

            # For each member model, look up the index
            for name, _, attr in lookup_members:
                model = attr(e)
                if model is not None:
                    obj[name] = index_lookup[name][model]

            result["experiment"].append(obj)

        def get_template(imset):
            if imset.reader().is_single_file_reader():
                return imset.reader().master_path()
            else:
                return imset.get_template()

        # Serialize all the imagesets
        result["imageset"] = []
        for imset in index_lookup["imageset"]:
            if isinstance(imset, ImageSequence):
                # FIXME_HACK
                template = get_template(imset)
                r = {
                    "__id__": "ImageSequence",
                    "template": template,
                }
                if imset.reader().is_single_file_reader():
                    r["single_file_indices"] = list(imset.indices())
            elif isinstance(imset, ImageSet):
                r = {
                    "__id__": "ImageSet",
                    "images": imset.paths(),
                }
                if imset.reader().is_single_file_reader():
                    r["single_file_indices"] = list(imset.indices())
            elif isinstance(imset, ImageGrid):
                r = {
                    "__id__": "ImageGrid",
                    "images": imset.paths(),
                    "grid_size": imset.get_grid_size(),
                }
                if imset.reader().is_single_file_reader():
                    r["single_file_indices"] = list(imset.indices())
            else:
                raise TypeError(
                    "expected ImageSet or ImageSequence, got %s" % type(imset)
                )
            r["mask"] = abspath_or_none(imset.external_lookup.mask.filename)
            r["gain"] = abspath_or_none(imset.external_lookup.gain.filename)
            r["pedestal"] = abspath_or_none(imset.external_lookup.pedestal.filename)
            r["dx"] = abspath_or_none(imset.external_lookup.dx.filename)
            r["dy"] = abspath_or_none(imset.external_lookup.dy.filename)
            r["params"] = imset.params()
            result["imageset"].append(r)

        # Extract all the ordered model dictionaries - is important these
        # preserve the same order as used in experiment serialization above
        for name, models in index_lookup.items():
            # Only fill out entries not handled above e.g. imageset
            if name not in result:
                result[name] = [x.to_dict() for x in models]

        return result

    def nullify_all_single_file_reader_format_instances(self):
        """
        Parallel reading of HDF5 from the same handle is not allowed. Python
        multiprocessing is a bit messed up and used fork on linux so need to
        close and reopen file.

        """
        for experiment in self:
            if experiment.imageset.reader().is_single_file_reader():
                experiment.imageset.reader().nullify_format_instance()

    def as_json(self, filename=None, compact=False, split=False):
        """Dump experiment list as json"""
        # Get the dictionary and get the JSON string
        dictionary = self.to_dict()

        # Split into separate files
        if filename is not None and split:
            # Get lists of models by filename
            basepath = os.path.splitext(filename)[0]
            ilist = [
                ("%s_imageset_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["imageset"])
            ]
            blist = [
                ("%s_beam_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["beam"])
            ]
            dlist = [
                ("%s_detector_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["detector"])
            ]
            glist = [
                ("%s_goniometer_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["goniometer"])
            ]
            slist = [
                ("%s_scan_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["scan"])
            ]
            clist = [
                ("%s_crystal_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["crystal"])
            ]
            plist = [
                ("%s_profile_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["profile"])
            ]
            scalelist = [
                ("%s_scaling_model_%d.json" % (basepath, i), d)
                for i, d in enumerate(dictionary["scaling_model"])
            ]

            # Get the list of experiments
            edict = {
                "__id__": "ExperimentList",
                "experiment": dictionary["experiment"],
            }

            # Set paths rather than indices
            for e in edict["experiment"]:
                if "imageset" in e:
                    e["imageset"] = ilist[e["imageset"]][0]
                if "beam" in e:
                    e["beam"] = blist[e["beam"]][0]
                if "detector" in e:
                    e["detector"] = dlist[e["detector"]][0]
                if "goniometer" in e:
                    e["goniometer"] = glist[e["goniometer"]][0]
                if "scan" in e:
                    e["scan"] = slist[e["scan"]][0]
                if "crystal" in e:
                    e["crystal"] = clist[e["crystal"]][0]
                if "profile" in e:
                    e["profile"] = plist[e["profile"]][0]
                if "scaling_model" in e:
                    e["scaling_model"] = scalelist[e["scaling_model"]][0]

            to_write = (
                ilist
                + blist
                + dlist
                + glist
                + slist
                + clist
                + plist
                + scalelist
                + [(filename, edict)]
            )
        else:
            to_write = [(filename, dictionary)]

        for fname, obj in to_write:
            if compact:
                separators = (",", ":")
                indent = None
            else:
                separators = None
                indent = 2
            text = json.dumps(
                obj,
                separators=separators,
                indent=indent,
                ensure_ascii=True,
                cls=AutoEncoder,
            )

            # If a filename is set then dump to file otherwise return string
            if fname:
                with open(fname, "w") as outfile:
                    outfile.write(text)
            else:
                return text

    def as_file(self, filename, **kwargs):
        """Dump experiment list as file."""
        ext = os.path.splitext(filename)[1]
        j_ext = [".json", ".expt"]
        if ext.lower() in j_ext:
            return self.as_json(filename, **kwargs)
        else:
            ext_str = "|".join(j_ext)
            raise RuntimeError(f"expected extension {{{ext_str}}}, got {ext}")

    @staticmethod
    def from_file(filename: str, check_format: bool = True) -> ExperimentList:
        """
        Load an ExperimentList from a serialized file.

        Args:
            filename: The filename to load an ExperimentList from
            check_format: If True, will attempt to verify image data type
        """
        # Inline to avoid recursive imports
        from .experiment_list import ExperimentListFactory

        return ExperimentListFactory.from_serialized_format(filename, check_format)

    def change_basis(self, change_of_basis_ops, in_place=False):
        """
        Apply change of basis operators to an ExperimentList

        Args:
            change_of_basis_ops: This can either be a single
                cctbx.sgtbx.change_of_basis_op to be applied to all experiments, or a
                list of operators, one per experiment.
            in_place (bool): Apply the change of basis operations in-place to the
                current ExperimentList. Default is to return a copy of the
                ExperimentList.

        Returns:
            The reindexed ExperimentList
        """
        if isinstance(change_of_basis_ops, cctbx.sgtbx.change_of_basis_op):
            change_of_basis_ops = [change_of_basis_ops] * len(self)
        assert len(change_of_basis_ops) == len(self), (
            "Number of change_of_basis_ops (%i) not equal to the number of experiments (%i)"
            % (len(change_of_basis_ops), len(self))
        )
        if in_place:
            return_expts = self
        else:
            return_expts = copy.deepcopy(self)
        for expt, cb_op in zip(return_expts, change_of_basis_ops):
            expt.crystal = expt.crystal.change_basis(cb_op)
        return return_expts

    @staticmethod
    def from_templates(templates, **kwargs):
        """Import an experiment list from templates"""
        # Import here to avoid cyclic dependencies
        from .experiment_list import ExperimentListFactory

        return ExperimentListFactory.from_templates(templates, **kwargs)



from __future__ import annotations

import sys

from boost_adaptbx.boost.python import streambuf
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.ext import (
    is_big_endian,
    read_uint8,
    read_uint16,
    read_uint16_bs,
    read_uint32,
    read_uint32_bs,
)
from dxtbx.format.FormatBruker import FormatBruker


class FormatBrukerPhoton(FormatBruker):
    @staticmethod
    def understand(image_file):

        try:
            header_lines = FormatBruker.read_header_lines(image_file)
        except OSError:
            return False

        header_dic = FormatBruker.parse_header(header_lines)

        dettype = header_dic.get("DETTYPE")
        if dettype is None:
            return False
        # We support Photon-II and Photon-III detectors
        if not dettype.startswith("CMOS-PHOTONII"):
            return False

        return True

    def _start(self):

        try:
            header_lines = FormatBruker.read_header_lines(self._image_file)
        except OSError:
            return False

        self.header_dict = FormatBrukerPhoton.parse_header(header_lines)

        # The Photon II/III format can't currently use BrukerImage, see
        # https://github.com/cctbx/cctbx_project/issues/65
        # from iotbx.detectors.bruker import BrukerImage
        # self.detectorbase = BrukerImage(self._image_file)

    def _goniometer(self):
        # goniometer angles in ANGLES are 2-theta, omega, phi, chi (FIXED)
        # AXIS indexes into this list to define the scan axis (in FORTRAN counting)
        # START and RANGE define the start and step size for each image

        _, omega, phi, chi = map(float, self.header_dict["ANGLES"].split())
        scan_axis = ["NONE", "2THETA", "OMEGA", "PHI", "CHI", "X", "Y", "Z"]
        scan_axis = scan_axis[int(self.header_dict["AXIS"])]
        names = flex.std_string(("PHI", "CHI", "OMEGA"))
        scan_axis = flex.first_index(names, scan_axis)
        if scan_axis is None:
            scan_axis = "OMEGA"  # default

        # https://journals.iucr.org/d/issues/2014/10/00/dz5309/dz5309sup1.pdf
        axes = flex.vec3_double(((0, -1, 0), (0, 0, 1), (0, 1, 0)))
        omega -= 180
        angles = flex.double((phi, chi, omega))

        return self._goniometer_factory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis
        )

    def _calculate_gain(self, wavelength):
        """The CCDPARM header item contains 5 items:
            1. readnoise
            2. e/ADU
            3. e/photon
            4. bias
            5. full scale
        The gain in ADU/X-ray is given by (e/photon) / (e/ADU).
        """
        ccdparm = self.header_dict["CCDPARM"].split()
        e_ADU = float(ccdparm[1])
        e_photon = float(ccdparm[2])
        if e_ADU == 0:
            return 1.0
        return e_photon / e_ADU

    def _detector(self):
        # goniometer angles in ANGLES are 2-theta, omega, phi, chi (FIXED)
        two_theta = float(self.header_dict["ANGLES"].split()[0])

        # Assume Bruker full_scale value means saturation
        full_scale = float(self.header_dict["CCDPARM"].split()[-1])
        min_trusted_value = 0

        fast = matrix.col((1, 0, 0))
        slow = matrix.col((0, 1, 0))
        beam = matrix.col((0, 0, 1))
        pixel_mm = 5.0 / float(self.header_dict["DETTYPE"].split()[1])
        beam_pixel = [float(bp) for bp in self.header_dict["CENTER"].split()[:-3:-1]]
        distance_mm = 10.0 * float(self.header_dict["DISTANC"].split()[1])
        origin = (
            -distance_mm * beam
            - fast * pixel_mm * beam_pixel[1]
            - slow * pixel_mm * beam_pixel[0]
        )
        # 2theta rotation appears to be around the slow axis
        origin = origin.rotate_around_origin(slow, two_theta, deg=True)
        fast = fast.rotate_around_origin(slow, two_theta, deg=True)
        pixel_size = pixel_mm, pixel_mm
        # ncols is nfast, nrows is nslow
        image_size = (
            int(self.header_dict["NCOLS"].split()[0]),
            int(self.header_dict["NROWS"].split()[0]),
        )

        # Not a CCD, but is an integrating detector. Photon II has a 90 um Gadox
        # scintillator.
        gain = self._calculate_gain(float(self.header_dict["WAVELEN"].split()[0]))
        return self._detector_factory.complex(
            "CCD",
            origin.elems,
            fast.elems,
            slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, full_scale),
            gain=gain,
        )

    def _beam(self):
        """Assume home source, so make unpolarized beam"""
        wavelength = float(self.header_dict["WAVELEN"].split()[0])

        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
        )

    def _scan(self):

        start = float(self.header_dict["START"].split()[0])
        incr = float(self.header_dict["INCREME"].split()[0])
        if incr < 0:
            start *= -1
            incr *= -1

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=1,
            osc_start=start,
            osc_width=incr,
            epoch=None,
        )

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        # It is better to catch FORMAT 86 here and fail with a sensible error msg
        # as soon as something is attempted with the image data rather than in
        # the understand method. Otherwise the user gets FormatBruker reading the
        # image improperly but without failing
        if self.header_dict["FORMAT"] != "100":
            raise NotImplementedError(
                "Only FORMAT 100 images from the Photon II are currently supported"
            )

        f = self.open_file(self._image_file, "rb")
        header_size = int(self.header_dict["HDRBLKS"]) * 512
        f.read(header_size)

        if is_big_endian():
            read_2b = read_uint16_bs
            read_4b = read_uint32_bs
        else:
            read_2b = read_uint16
            read_4b = read_uint32

        # NPIXELB stores the number of bytes/pixel for the data and the underflow
        # table. We expect 1 byte for underflows and either 2 or 1 byte per pixel
        # for the data
        npixelb = [int(e) for e in self.header_dict["NPIXELB"].split()]
        assert npixelb[1] == 1

        if npixelb[0] == 1:
            read_data = read_uint8
        elif npixelb[0] == 2:
            read_data = read_2b
        else:
            raise IncorrectFormatError(
                "{} bytes per pixel is not supported".format(npixelb[0])
            )

        nrows = int(self.header_dict["NROWS"].split()[0])
        ncols = int(self.header_dict["NCOLS"].split()[0])

        raw_data = read_data(streambuf(f), nrows * ncols)

        image_size = (nrows, ncols)
        raw_data.reshape(flex.grid(*image_size))

        (num_underflows, num_2b_overflows, num_4b_overflows) = [
            int(e) for e in self.header_dict["NOVERFL"].split()
        ]

        # read underflows
        if num_underflows > 0:
            # stored values are padded to a multiple of 16 bytes
            nbytes = num_underflows + 15 & ~(15)
            underflow_vals = read_uint8(streambuf(f), nbytes)[:num_underflows]
        else:
            underflow_vals = None

        # handle 2 byte overflows
        if num_2b_overflows > 0:
            # stored values are padded to a multiple of 16 bytes
            nbytes = num_2b_overflows * 2 + 15 & ~(15)
            overflow_vals = read_2b(streambuf(f), nbytes // 2)[:num_2b_overflows]
            overflow = flex.int(nrows * ncols, 0)
            sel = (raw_data == 255).as_1d()
            overflow.set_selected(sel, overflow_vals - 255)
            overflow.reshape(flex.grid(*image_size))
            raw_data += overflow

        # handle 4 byte overflows
        if num_4b_overflows > 0:
            # stored values are padded to a multiple of 16 bytes
            nbytes = num_4b_overflows * 4 + 15 & ~(15)
            overflow_vals = read_4b(streambuf(f), nbytes // 4)[:num_4b_overflows]
            overflow = flex.int(nrows * ncols, 0)
            sel = (raw_data == 65535).as_1d()
            overflow.set_selected(sel, overflow_vals - 65535)
            overflow.reshape(flex.grid(*image_size))
            raw_data += overflow

        # handle underflows
        if underflow_vals is not None:
            sel = (raw_data == 0).as_1d()
            underflow = flex.int(nrows * ncols, 0)
            underflow.set_selected(sel, underflow_vals)
            underflow.reshape(flex.grid(*image_size))
            raw_data += underflow

        # handle baseline. num_underflows == -1 means no baseline subtraction. See
        # https://github.com/cctbx/cctbx_project/files/1262952/BISFrameFileFormats.zip
        if num_underflows != -1:
            num_exposures = [int(e) for e in self.header_dict["NEXP"].split()]
            baseline = num_exposures[2]
            raw_data += baseline

        return raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatBrukerPhoton.understand(arg))



"""
An implementation of the SMV image reader for Rigaku Saturn images, for
the instrument at Dundee SN 11480296. Inherits from
FormatSMVRigakuSaturn customizing only the difference (incorrect definition
of fast and slow directions.)
"""

from __future__ import annotations

import sys

from scitbx import matrix

from dxtbx.format.FormatSMVRigakuSaturn import FormatSMVRigakuSaturn


class FormatSMVRigakuSaturnSN11480296(FormatSMVRigakuSaturn):
    """A class for reading SMV format Rigaku Saturn images, and correctly
    constructing a model for the experiment from this. This is the custom
    version for instrument with serial number 11480296 at Dundee."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku Saturn SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for. Also checks the serial number."""

        size, header = FormatSMVRigakuSaturn.get_smv_header(image_file)

        # FIXME if they fix this "bug" in the header sometime then may need
        # to add in an extra check that the data were recorded before this
        # event.

        detector_prefix = header["DETECTOR_NAMES"].split()[0].strip()
        serial_number = header.get("%sSERIAL_NUMBER" % detector_prefix)

        return serial_number == "11480296"

    def _detector(self):
        """Return a model for the detector, allowing for two-theta offsets
        and the detector position. This will be rather more complex... and
        overloads the definition for the general Rigaku Saturn detector by
        rotating the detector by -90 degrees about the beam."""

        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        R = matrix.col((0, 0, 1)).axis_and_angle_as_r3_rotation_matrix(-90, deg=True)
        detector_fast = R * matrix.col(tuple(detector_axes[:3]))
        detector_slow = R * matrix.col(tuple(detector_axes[3:]))

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(self._header_dictionary["SATURATED_VALUE"])
        min_trusted_value = 0

        return self._detector_factory.complex(
            "CCD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
            gain=10,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVRigakuSaturnSN11480296.understand(arg))



from __future__ import annotations

import copy

import pytest

from libtbx.phil import parse

from dxtbx.model import Scan
from dxtbx.model.scan import ScanFactory, scan_phil_scope


@pytest.fixture
def scan():
    image_range = (0, 1000)
    oscillation = (0, 0.1)
    scan = Scan(image_range, oscillation)
    scan.set_batch_offset(100)
    return scan


def test_is_batch_valid(scan):
    """Check that the is_batch_valid function behaves properly."""
    br1, br2 = scan.get_batch_range()
    for i in range(0, br1):
        assert scan.is_batch_valid(i) is False
    for i in range(br1, br2 + 1):
        assert scan.is_batch_valid(i) is True
    for i in range(br2 + 1, br2 + 100):
        assert scan.is_batch_valid(i) is False


def test_batch_deepcopy(scan):
    """Check that copy.deepcopy(scan) behaves properly."""
    assert scan.get_batch_offset() == 100
    scan1 = copy.deepcopy(scan)
    assert scan1.get_batch_offset() == scan.get_batch_offset()


def test_scan_slice(scan):
    assert scan.get_batch_offset() == 100
    scan1 = scan[:10]
    assert scan1.get_batch_offset() == scan.get_batch_offset()


def test_is_angle_valid(scan):
    """Check that the is_angle_valid function behaves properly."""
    oscillation_range = scan.get_oscillation_range()
    os1 = int(oscillation_range[0])
    os2 = int(oscillation_range[1])
    for i in range(0, os1):
        assert scan.is_angle_valid(i) is False
    for i in range(os1, os2):
        assert scan.is_angle_valid(i) is True
    for i in range(os2 + 1, 360):
        assert scan.is_angle_valid(i) is False


def test_is_frame_valid(scan):
    """Check that the is_frame_valid function behaves properly."""
    image_range = scan.get_image_range()
    for i in range(image_range[0] - 100, image_range[0]):
        assert scan.is_image_index_valid(i) is False
    for i in range(image_range[0], image_range[1] + 1):
        assert scan.is_image_index_valid(i) is True
    for i in range(image_range[1] + 1, image_range[1] + 100):
        assert scan.is_image_index_valid(i) is False


def test_scan_oscillation_recycle(scan):
    for deg in (True, False):
        oscillation = scan.get_oscillation(deg=deg)
        scan.set_oscillation(oscillation, deg=deg)
        assert scan.get_oscillation(deg=deg) == oscillation


def test_scan_360_append():
    scan1 = Scan((1, 360), (0.0, 1.0))
    scan2 = Scan((361, 720), (0.0, 1.0))

    scan = scan1 + scan2
    eps = 1e-7
    assert scan.get_num_images() == 720
    assert abs(scan.get_oscillation()[0] - 0.0) < eps
    assert abs(scan.get_oscillation()[1] - 1.0) < eps
    assert scan.get_image_range() == (1, 720)
    assert scan.get_batch_range() == (1, 720)

    scan1 = Scan((1, 360), (0.0, 1.0))
    scan2 = Scan((361, 720), (360.0, 1.0))

    scan = scan1 + scan2
    eps = 1e-7
    assert scan.get_num_images() == 720
    assert abs(scan.get_oscillation()[0] - 0.0) < eps
    assert abs(scan.get_oscillation()[1] - 1.0) < eps
    assert scan.get_image_range() == (1, 720)
    assert scan.get_batch_range() == (1, 720)

    scan2.set_batch_offset(10)
    with pytest.raises(RuntimeError):
        scan = scan1 + scan2


def test_swap():
    scan1 = Scan((1, 20), (0.0, 1.0))
    scan2 = Scan((40, 60), (10.0, 2.0))
    scan1.swap(scan2)
    assert scan2.get_image_range() == (1, 20)
    assert scan1.get_image_range() == (40, 60)
    assert scan2.get_oscillation() == (0.0, 1.0)
    assert scan1.get_oscillation() == (10.0, 2.0)


def test_valid_image_ranges():
    scan = Scan((1, 100), (0.0, 1.0))
    scan.set_valid_image_ranges("0", [(1, 80)])
    scan.set_valid_image_ranges("1", [(1, 50)])
    assert list(scan.get_valid_image_ranges("0")) == [(1, 80)]
    assert list(scan.get_valid_image_ranges("1")) == [(1, 50)]
    scan.set_valid_image_ranges("0", [(1, 50), (80, 100)])
    assert list(scan.get_valid_image_ranges("0")) == [(1, 50), (80, 100)]
    assert list(scan.get_valid_image_ranges("2")) == []

    assert scan.get_image_range() == (1, 100)
    scan2 = copy.copy(scan)
    assert scan2.get_image_range() == (1, 100)
    assert list(scan2.get_valid_image_ranges("0")) == [(1, 50), (80, 100)]


def test_from_phil():
    params = scan_phil_scope.fetch(
        parse(
            """
    scan {
      image_range = 1, 10
      oscillation = (-4, 0.1)
    }
  """
        )
    ).extract()

    s1 = ScanFactory.from_phil(params)

    assert s1.get_num_images() == 10
    assert s1.get_image_range() == (1, 10)
    assert s1.get_oscillation() == (-4, 0.1)
    assert s1.get_batch_offset() == 0
    assert s1.get_batch_range() == s1.get_image_range()
    for i in range(s1.get_image_range()[0], s1.get_image_range()[1] + 1):
        assert s1.get_batch_for_image_index(i) == i
        assert s1.get_batch_for_array_index(i - 1) == i

    params = scan_phil_scope.fetch(
        parse(
            """
    scan {
      image_range = 1, 20
      extrapolate_scan = True
      oscillation = (20, 0.01)
      batch_offset = 10
    }
  """
        )
    ).extract()

    s2 = ScanFactory.from_phil(params, s1)
    assert s2.get_num_images() == 20
    assert s2.get_image_range() == (1, 20)
    assert s2.get_oscillation() == (20, 0.01)
    assert s2.get_batch_offset() == 10
    assert s2.get_batch_range() == (11, 30)
    ir1, ir2 = s2.get_image_range()
    for i in range(ir1, ir2):
        assert s2.get_batch_for_image_index(i) == i + s2.get_batch_offset()
        assert s2.is_batch_valid(s2.get_batch_for_image_index(i))


def test_scan_factory_from_dict(scan):

    empty_scan = Scan()
    empty_scan_from_dict = ScanFactory.from_dict(empty_scan.to_dict())

    assert empty_scan_from_dict == empty_scan

    scan_from_dict = ScanFactory.from_dict(scan.to_dict())

    assert scan_from_dict == scan



from __future__ import annotations

import pytest

from dxtbx.sequence_filenames import template_regex


@pytest.mark.parametrize(
    "filename,template,digits",
    [
        ("foo_bar_001.img", "foo_bar_###.img", 1),
        ("foo_bar001.img", "foo_bar###.img", 1),
        ("foo_bar_1.8A_004.img", "foo_bar_1.8A_###.img", 4),
        ("foo_bar.001", "foo_bar.###", 1),
        ("foo_bar_002.img1000", "foo_bar_###.img1000", 2),
        ("foo_bar_00005.img", "foo_bar_#####.img", 5),
        ("image0010", "image####", 10),
    ],
)
def test_template_regex(filename, template, digits):
    assert template_regex(filename) == (template, digits)



"""
Note, scans and gonios not supported here. This writer essentially writes still images

Example to write the first 10 images from an h5 file:
writer = FullCBFWriter("data.h5")
for i in range(10):
  writer.write_cbf("example_%d.cbf"%i, index=i)
"""

from __future__ import annotations

import os
import sys

import pycbf
from serialtbx.detector import basis

from scitbx.array_family import flex

import dxtbx.format.Registry
from dxtbx.format.FormatCBFMultiTile import cbf_wrapper


def add_frame_specific_cbf_tables(
    cbf,
    wavelength,
    timestamp,
    trusted_ranges,
    diffrn_id="DS1",
    is_xfel=True,
    gain=1.0,
    flux=None,
):
    """Adds tables to cbf handle that won't already exsist if the cbf file is just a header
    @ param wavelength Wavelength in angstroms
    @ param timestamp String formatted timestamp for the image
    @ param trusted_ranges Array of trusted range tuples (min, max), one for each element"""

    """Data items in the DIFFRN_RADIATION category describe
   the radiation used for measuring diffraction intensities,
   its collimation and monochromatization before the sample.

   Post-sample treatment of the beam is described by data
   items in the DIFFRN_DETECTOR category."""
    if flux:
        cbf.add_category(
            "diffrn_radiation", ["diffrn_id", "wavelength_id", "probe", "beam_flux"]
        )
        cbf.add_row([diffrn_id, "WAVELENGTH1", "x-ray", "%f" % flux])
    else:
        cbf.add_category("diffrn_radiation", ["diffrn_id", "wavelength_id", "probe"])
        cbf.add_row([diffrn_id, "WAVELENGTH1", "x-ray"])

    """ Data items in the DIFFRN_RADIATION_WAVELENGTH category describe
   the wavelength of the radiation used in measuring the diffraction
   intensities. Items may be looped to identify and assign weights
   to distinct wavelength components from a polychromatic beam."""
    cbf.add_category("diffrn_radiation_wavelength", ["id", "wavelength", "wt"])
    cbf.add_row(["WAVELENGTH1", str(wavelength), "1.0"])

    """Data items in the DIFFRN_MEASUREMENT category record details
   about the device used to orient and/or position the crystal
   during data measurement and the manner in which the
   diffraction data were measured."""
    cbf.add_category(
        "diffrn_measurement", ["diffrn_id", "id", "number_of_axes", "method", "details"]
    )
    cbf.add_row(
        [
            diffrn_id,
            "INJECTION" if is_xfel else "unknown",
            "0",
            "electrospray"
            if is_xfel
            else "unknown" "crystals injected by electrospray"
            if is_xfel
            else "unknown",
        ]
    )

    """ Data items in the DIFFRN_SCAN category describe the parameters of one
     or more scans, relating axis positions to frames."""
    cbf.add_category("diffrn_scan", ["id", "frame_id_start", "frame_id_end", "frames"])
    cbf.add_row(["SCAN1", "FRAME1", "FRAME1", "1"])

    """Data items in the DIFFRN_SCAN_FRAME category describe
   the relationships of particular frames to scans."""
    cbf.add_category(
        "diffrn_scan_frame",
        ["frame_id", "frame_number", "integration_time", "scan_id", "date"],
    )
    cbf.add_row(["FRAME1", "1", "0.0", "SCAN1", timestamp])

    """ Data items in the ARRAY_INTENSITIES category record the
   information required to recover the intensity data from
   the set of data values stored in the ARRAY_DATA category."""
    # More detail here: http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Carray_intensities.html
    array_names = []
    cbf.find_category(b"diffrn_data_frame")
    while True:
        try:
            cbf.find_column(b"array_id")
            array_names.append(cbf.get_value().decode())
            cbf.next_row()
        except Exception as e:
            assert "CBF_NOTFOUND" in str(e)
            break

    if not isinstance(gain, list):
        gain = [gain] * len(array_names)

    cbf.add_category(
        "array_intensities",
        [
            "array_id",
            "binary_id",
            "linearity",
            "gain",
            "gain_esd",
            "overload",
            "underload",
            "undefined_value",
        ],
    )
    for i, array_name in enumerate(array_names):
        overload = trusted_ranges[i][1] + 1
        underload = trusted_ranges[i][0]
        undefined = underload - 1
        cbf.add_row(
            [
                array_name,
                str(i + 1),
                "linear",
                "%f" % gain[i],
                "0.0",
                str(overload),
                str(underload),
                str(undefined),
            ]
        )


class FullCBFWriter:
    """Class for writing full CBF files from any dxtbx-supported format class"""

    def __init__(self, filename=None, imageset=None):
        """Provide a file name or imageset as input"""
        assert [filename, imageset].count(
            None
        ) == 1, "Supply either filename or imageset"

        if filename is not None:
            format_class = dxtbx.format.Registry.get_format_class_for_file(filename)
            imageset = format_class.get_imageset([filename])

        self.imageset = imageset

    def get_metrology_dict(self, index=None):
        """Build a metrology dictionary.  This dictionary maps hierarchy keys to basis
        objects. A hierarchy key looks like this (0,1,2), where the entries are
        levels in a hierarchy and the numbers refer to a panel or group within that
        level"""
        metro = {}

        def recursive_setup_dict(panelgroup, key):
            metro[key] = basis(panelgroup=panelgroup)
            if panelgroup.is_panel():
                return
            for i, child in enumerate(panelgroup):
                childkey = tuple(list(key) + [i])
                recursive_setup_dict(child, childkey)

        if index is None:
            detector = self.imageset.get_detector()
        else:
            detector = self.imageset.get_detector(index)

        recursive_setup_dict(detector.hierarchy(), (0,))
        return metro

    def get_cbf_handle(self, index=None, header_only=False, detector_only=False):
        """Build a cbf handle in memory"""
        # set up the metrology dictionary to include axis names, pixel sizes, and so forth
        if index is None:
            detector = self.imageset.get_detector()
            beam = self.imageset.get_beam()
        else:
            detector = self.imageset.get_detector(index)
            beam = self.imageset.get_beam(index)

        metro = self.get_metrology_dict()

        def panel_group_from_key(key):
            # Find the node that a hierarchy key refers to
            if len(key) == 1:
                assert key[0] == 0
                return detector.hierarchy()

            node = detector.hierarchy()
            for i in key[1:]:
                node = node[i]
            return node

        def level_string(key):
            # Example for key (0,1,2). "L0M0_L1M1_L2M2", where L is level and M is module
            return "_".join(["L%dM%d" % (l, m) for l, m in enumerate(key)])

        # set up the metrology dictionary to include axis names, pixel sizes, and so forth
        dname = None
        detector_axes_names = []  # save these for later
        panelkeys = []
        panelnames = []

        def recursive_setup_basis_dict(key, parent_name="", panel_id=0):
            # Set up CBF axis names, including equipment components and depends_on chains
            basis = metro[key]
            node = panel_group_from_key(key)
            nodename = level_string(key)
            if basis.name:
                dname = basis.name
            else:
                dname = "AXIS_" + nodename

            if node.is_panel():
                panelname = "PANEL_%d" % panel_id
                panelkeys.append(key)
                panelnames.append(panelname)
                panel_id += 1

            if len(key) == 1:
                assert key == (0,)  # only one detector allowed for now

                for a in ["_X", "_Y", "_Z", "_R"]:
                    detector_axes_names.append(dname + a)
                basis.depends_on = dname + "_X"
                dname += "_R"
            else:
                detector_axes_names.append(dname)
                basis.depends_on = parent_name

            basis.equipment_component = "detector_level_%d" % (len(key) - 1)
            basis.axis_name = detector_axes_names[-1]
            if not node.is_panel():
                for c, child in enumerate(node):
                    panel_id = recursive_setup_basis_dict(
                        tuple(list(key) + [c]), dname, panel_id
                    )
            return panel_id

        recursive_setup_basis_dict((0,))

        if index is None:
            cbf_root = self.imageset.paths()[0]
        else:
            cbf_root = self.imageset.paths()[index]
        cbf_root = os.path.splitext(os.path.basename(cbf_root))[0]

        # the data block is the root cbf node
        cbf = cbf_wrapper()
        cbf.new_datablock(cbf_root.encode())

        # Each category listed here is preceded by the imageCIF description taken from here:
        # http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/index.html

        """Data items in the DIFFRN category record details about the
     diffraction data and their measurement."""
        diffrn_id = "dxtbx_%s" % self.imageset.reader().format_class.__name__
        cbf.add_category("diffrn", ["id"])
        cbf.add_row([diffrn_id])

        """Data items in the DIFFRN_SOURCE category record details of
    the source of radiation used in the diffraction experiment."""
        cbf.add_category("diffrn_source", ["diffrn_id", "source", "type"])
        cbf.add_row([diffrn_id, "unknown", "unknown"])

        """Data items in the DIFFRN_DETECTOR category describe the
     detector used to measure the scattered radiation, including
     any analyser and post-sample collimation."""
        cbf.add_category(
            "diffrn_detector", ["diffrn_id", "id", "type", "details", "number_of_axes"]
        )
        detector_id = "dxtbx_detector"
        cbf.add_row(
            [
                diffrn_id,
                detector_id,
                "General dxtbx detector",
                ".",
                str(len(detector_axes_names)),
            ]
        )

        """Data items in the DIFFRN_DETECTOR_AXIS category associate
       axes with detectors."""
        # Note, does not include the fast and the slow axes
        cbf.add_category("diffrn_detector_axis", ["detector_id", "axis_id"])
        for name in detector_axes_names:
            cbf.add_row([detector_id, name])

        """Data items in the DIFFRN_DETECTOR_ELEMENT category record
     the details about spatial layout and other characteristics
     of each element of a detector which may have multiple elements."""
        cbf.add_category("diffrn_detector_element", ["id", "detector_id"])

        for panelname in panelnames:
            cbf.add_row(["ELE_" + panelname, detector_id])

        """Data items in the DIFFRN_DATA_FRAME category record
     the details about each frame of data."""
        cbf.add_category(
            "diffrn_data_frame", ["id", "detector_element_id", "array_id", "binary_id"]
        )

        for i, panelname in enumerate(panelnames):
            cbf.add_row(
                ["FRAME1", "ELE_" + panelname, "ARRAY_" + panelname, "%d" % (i + 1)]
            )

        if not detector_only:
            # add_frame_specific_cbf_tables creates _array_intensities.undefined_value
            # and _array_intensities.overload. The trusted_range of the detector is
            # defined as [min-trusted-value, max-trusted-value]. The CBF definition
            # of 'overload' is in fact saturation - i.e. the max-trusted-value, while
            # the undefined_value is below the min-trusted-value.
            trusted_ranges = [panel.get_trusted_range() for panel in detector]
            try:
                add_frame_specific_cbf_tables(
                    cbf,
                    beam.get_wavelength(),
                    "unknown",
                    trusted_ranges,
                    diffrn_id,
                    False,
                    gain=[panel.get_gain() for panel in detector],
                    flux=beam.get_flux(),
                )
            except TypeError:
                # Needed until next cctbx release
                add_frame_specific_cbf_tables(
                    cbf,
                    beam.get_wavelength(),
                    "unknown",
                    trusted_ranges,
                    diffrn_id,
                    False,
                    gain=[panel.get_gain() for panel in detector],
                )

        """Data items in the AXIS category record the information required
       to describe the various goniometer, detector, source and other
       axes needed to specify a data collection.  The location of each
       axis is specified by two vectors: the axis itself, given as a unit
       vector, and an offset to the base of the unit vector.  These vectors
       are referenced to a right-handed laboratory coordinate system with
       its origin in the sample or specimen"""
        # More detail here: http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Caxis.html
        # Note we also use two new columns not in the latest imageCIF dictionary: rotation and rotation_axis.
        # We use them to specify an translation and a rotation in a single axis to describe a change in setting
        # when moving from one frame (say a quadrant) to another (say a sensor)
        cbf.add_category(
            "axis",
            [
                "id",
                "type",
                "equipment",
                "depends_on",
                "vector[1]",
                "vector[2]",
                "vector[3]",
                "offset[1]",
                "offset[2]",
                "offset[3]",
                "equipment_component",
            ],
        )

        # Keep a list of rows to add to the scan frame axis table
        axis_settings = []
        # keep a list of rows to add to the scan axis table
        axis_names = []

        root_key = (0,)
        dname = metro[root_key].axis_name[:-2]
        eqc = metro[root_key].equipment_component

        # Create a series of axis describing frame shifts from each level of the detector to the next
        cbf.add_row(
            "AXIS_SOURCE  general     source   .        0  0  1 . . . .".split()
        )
        axis_names.append("AXIS_SOURCE")
        cbf.add_row(
            "AXIS_GRAVITY general     gravity  .        0 -1  0 . . . .".split()
        )
        axis_names.append("AXIS_GRAVITY")
        cbf.add_row(
            (
                "%s_Z         translation detector .        0  0  1 . . . %s"
                % (dname, eqc)
            ).split()
        )
        axis_names.append("%s_Z" % dname)
        cbf.add_row(
            (
                "%s_Y         translation detector %s_Z     0  1  0 . . . %s"
                % (dname, dname, eqc)
            ).split()
        )
        axis_names.append("%s_Y" % dname)
        cbf.add_row(
            (
                "%s_X         translation detector %s_Y     1  0  0 . . . %s"
                % (dname, dname, eqc)
            ).split()
        )
        axis_names.append("%s_X" % dname)

        axis_settings.append(["AXIS_SOURCE", "FRAME1", "0", "0"])
        axis_settings.append(["AXIS_GRAVITY", "FRAME1", "0", "0"])
        axis_settings.append([dname + "_X", "FRAME1", "0", "0"])
        axis_settings.append([dname + "_Y", "FRAME1", "0", "0"])
        axis_settings.append([dname + "_Z", "FRAME1", "0", "0"])

        for key in sorted(metro):
            basis = metro[key]

            cbf.add_frame_shift(basis, axis_settings)
            axis_names.append(basis.axis_name)

            node = panel_group_from_key(key)

            if node.is_panel():
                axis_settings[-1][
                    -2
                ] = "0"  # Drop the setting change for leaves as it's encoded below

                aname = level_string(key)
                fast = [str(v) for v in node.get_local_fast_axis()]
                slow = [str(v) for v in node.get_local_slow_axis()]

                cbf.add_row(
                    [
                        "AXIS_" + aname + "_S",
                        "translation",
                        "detector",
                        basis.axis_name,
                        slow[0],
                        slow[1],
                        slow[2],
                        "0",
                        "0",
                        "0",
                        basis.equipment_component,
                    ]
                )
                cbf.add_row(
                    [
                        "AXIS_" + aname + "_F",
                        "translation",
                        "detector",
                        "AXIS_" + aname + "_S",
                        fast[0],
                        fast[1],
                        fast[2],
                        "0",
                        "0",
                        "0",
                        basis.equipment_component,
                    ]
                )
                axis_names.append("AXIS_" + aname + "_F")
                axis_names.append("AXIS_" + aname + "_S")
                axis_settings.append(["AXIS_" + aname + "_F", "FRAME1", "0", "0"])
                axis_settings.append(["AXIS_" + aname + "_S", "FRAME1", "0", "0"])

        """Data items in the DIFFRN_SCAN_AXIS category describe the settings of
       axes for particular scans.  Unspecified axes are assumed to be at
       their zero points."""
        # leave all the settings zero. Levels with settings are set below.
        cbf.add_category(
            "diffrn_scan_axis",
            [
                "axis_id",
                "scan_id",
                "angle_start",
                "angle_range",
                "angle_increment",
                "displacement_start",
                "displacement_range",
                "displacement_increment",
            ],
        )
        for name in axis_names:
            cbf.add_row([name, "SCAN1", "0", "0", "0", "0", "0", "0"])

        """Data items in the DIFFRN_SCAN_FRAME_AXIS category describe the
       settings of axes for particular frames.  Unspecified axes are
       assumed to be at their zero points."""
        cbf.add_category(
            "diffrn_scan_frame_axis", ["axis_id", "frame_id", "angle", "displacement"]
        )
        for row in axis_settings:
            cbf.add_row(row)

        """Data items in the ARRAY_STRUCTURE_LIST category record the size
       and organization of each array dimension.
       The relationship to physical axes may be given."""

        cbf.add_category(
            "array_structure_list",
            [
                "array_id",
                "index",
                "dimension",
                "precedence",
                "direction",
                "axis_set_id",
            ],
        )
        for panelkey, panelname in zip(panelkeys, panelnames):
            aname = level_string(panelkey)
            panel = panel_group_from_key(panelkey)
            fast_dim, slow_dim = panel.get_image_size()
            cbf.add_row(
                [
                    "ARRAY_" + panelname,
                    "1",
                    "%d" % fast_dim,
                    "1",
                    "increasing",
                    "AXIS_" + aname + "_F",
                ]
            )
            cbf.add_row(
                [
                    "ARRAY_" + panelname,
                    "2",
                    "%d" % slow_dim,
                    "2",
                    "increasing",
                    "AXIS_" + aname + "_S",
                ]
            )

        """Data items in the ARRAY_STRUCTURE_LIST_SECTION category identify
       the dimension-by-dimension start, end and stride of each section of an
       array that is to be referenced."""
        # no sections here

        """Data items in the ARRAY_STRUCTURE_LIST_AXIS category describe
       the physical settings of sets of axes for the centres of pixels that
       correspond to data points described in the
       ARRAY_STRUCTURE_LIST category."""
        cbf.add_category(
            "array_structure_list_axis",
            ["axis_set_id", "axis_id", "displacement", "displacement_increment"],
        )
        for panelkey in panelkeys:
            aname = level_string(panelkey)
            panel = panel_group_from_key(panelkey)
            pixel_size = panel.get_pixel_size()
            cbf.add_row(
                [
                    "AXIS_" + aname + "_F",
                    "AXIS_" + aname + "_F",
                    "0.0",
                    str(pixel_size[0]),
                ]
            )
            cbf.add_row(
                [
                    "AXIS_" + aname + "_S",
                    "AXIS_" + aname + "_S",
                    "0.0",
                    str(pixel_size[1]),
                ]
            )

        if not header_only:
            self.add_data_to_cbf(cbf, index)

        return cbf

    def add_data_to_cbf(self, cbf, index=None, data=None, verbose=False):
        """
        Given a cbf handle, add the raw data and the necessary tables to support it
        """
        if data is None:
            if index is None:
                data = self.imageset.get_raw_data(0)
            else:
                data = self.imageset.get_raw_data(index)
        if not isinstance(data, tuple):
            data = (data,)

        array_names = []
        cbf.find_category(b"diffrn_data_frame")
        while True:
            try:
                cbf.find_column(b"array_id")
                array_names.append(cbf.get_value())
                cbf.next_row()
            except Exception as e:
                e_message = str(e)

                assert "CBF_NOTFOUND" in e_message
                break

        dataisint = flex.bool()
        for panel_data in data:
            assert len(panel_data.focus()) == 2
            if isinstance(panel_data, flex.int):
                dataisint.append(True)
            elif isinstance(panel_data, flex.double):
                dataisint.append(False)
            else:
                raise TypeError("Ints or doubles are required")

        """ Data items in the ARRAY_STRUCTURE category record the organization and
    encoding of array data in the ARRAY_DATA category."""
        cbf.add_category(
            "array_structure", ["id", "encoding_type", "compression_type", "byte_order"]
        )
        for i, array_name in enumerate(array_names):
            if dataisint[i]:
                cbf.add_row(
                    [array_name, "signed 32-bit integer", "packed", "little_endian"]
                )
            else:
                cbf.add_row(
                    [array_name, "signed 64-bit real IEEE", "packed", "little_endian"]
                )

        """ Data items in the ARRAY_DATA category are the containers for the array data
    items described in the category ARRAY_STRUCTURE. """
        cbf.add_category("array_data", ["array_id", "binary_id", "data"])

        if verbose:
            print("Compressing tiles...")

        for i, (panel_data, array_name) in enumerate(zip(data, array_names)):
            focus = panel_data.focus()
            # panel_data += 1

            cbf.add_row([array_name, str(i + 1)])

            binary_id = i + 1
            byte_str = panel_data.copy_to_byte_str()
            elements = len(panel_data)
            byteorder = "little_endian"
            dimfast = focus[1]
            dimmid = focus[0]
            dimslow = 1
            padding = 0

            if dataisint[i]:
                elsize = 4
                elsigned = 1
                byteorder = byteorder.encode()
                cbf.set_integerarray_wdims_fs(
                    pycbf.CBF_PACKED,
                    binary_id,
                    byte_str,
                    elsize,
                    elsigned,
                    elements,
                    byteorder,
                    dimfast,
                    dimmid,
                    dimslow,
                    padding,
                )
            else:
                elsize = 8
                byteorder = byteorder.encode()
                cbf.set_realarray_wdims_fs(
                    pycbf.CBF_PACKED,
                    binary_id,
                    byte_str,
                    elsize,
                    elements,
                    byteorder,
                    dimfast,
                    dimmid,
                    dimslow,
                    padding,
                )

    def write_cbf(self, filename, index=None, cbf=None):
        """Write a CBF file. If the handle is not provided, create one"""
        assert [index, cbf].count(None) in (1, 2), "Supply either index or cbf"

        if cbf is None:
            cbf = self.get_cbf_handle(index=index, header_only=True)
            self.add_data_to_cbf(cbf, index=index)

        filename = filename.encode()
        cbf.write_widefile(
            filename, pycbf.CBF, pycbf.MIME_HEADERS | pycbf.MSG_DIGEST | pycbf.PAD_4K, 0
        )


if __name__ == "__main__":
    filename = sys.argv[1]
    if len(sys.argv) > 2:
        index: int | None = int(sys.argv[2])
    else:
        index = None

    writer = FullCBFWriter(filename=filename)
    writer.write_cbf("converted.cbf", index)



from __future__ import annotations

import argparse

import dxtbx.util
from dxtbx.sequence_filenames import find_matching_images


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(
        description="Find images that match a template specification"
    )
    parser.add_argument(
        "template", help="The template specification", metavar="TEMPLATE"
    )
    options = parser.parse_args(args)

    for mi in find_matching_images(options.template):
        print(mi)


if __name__ == "__main__":
    run()



from __future__ import annotations

import sys

from dxtbx.format.FormatSMVADSC import FormatSMVADSC


class FormatSMVADSCmlfsom(FormatSMVADSC):
    """A class for reading SMV::ADSC-format images generated by MLFSOM
    simulation."""

    @staticmethod
    def understand(image_file):

        size, header = FormatSMVADSC.get_smv_header(image_file)

        unwanted_header_items = ["TIME", "DATE"]

        for header_item in unwanted_header_items:
            if header_item in header:
                return False

        return True

    def _scan(self):
        """Return the scan information for this image."""

        exposure_time = 1.0  # dummy argument; ought to be explicitly output by MLFSOM!
        epoch = None

        # assert(epoch)
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCmlfsom.understand(arg))



from __future__ import annotations

import pickle
import sys
from calendar import timegm
from time import strptime

from serialtbx.detector.legacy_metrology.cspad_detector import CSPadDetector

from iotbx.detectors.npy import image_dict_to_unicode
from scitbx.matrix import col

from dxtbx.format.FormatPY import FormatPY
from dxtbx.model import Detector


class FormatPYmultitile(FormatPY):
    @staticmethod
    def understand(image_file):
        try:
            with FormatPYmultitile.open_file(image_file, "rb") as fh:
                data = pickle.load(fh, encoding="bytes")
                data = image_dict_to_unicode(data)
        except OSError:
            return False

        wanted_header_items = ["TILES", "METROLOGY"]
        if any(header_item not in data for header_item in wanted_header_items):
            return False

        unwanted_header_items = ["SIZE1", "SIZE2"]
        if any(header_item in data for header_item in unwanted_header_items):
            return False

        return True

    def _start(self):
        # this Format class depends on stuff from detectorbase
        self.detectorbase_start()

    def detectorbase_start(self):
        self.detectorbase = CSPadDetector(self._image_file)
        self.detectorbase.readHeader()
        self._metrology_params = self.detectorbase._metrology_params
        self._tiles = self.detectorbase._tiles

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """The _detector() function returns a model for a CSPAD detector as
        used at LCLS's CXI and XPP endstations.  It converts the
        metrology information in the pure Python object extracted from
        the image pickle to DXTBX-style transformation vectors.  Only
        ASIC:s are considered, since DXTBX metrology is not concerned
        with hierarchies.

        Merged from serialtbx.detector.legacy_metrology.cspad_detector.readHeader()
        and serialtbx.detector.legacy_metrology.metrology.metrology_as_dxtbx_vectors().
        """

        from serialtbx.detector.legacy_metrology.metrology import (
            _transform,
            get_projection_matrix,
        )

        # Apply the detector distance to the translation of the root
        # detector object.
        d = self._metrology_params.detector
        Tb_d = _transform(
            col(d.orientation).normalize(),
            col(d.translation) + col((0, 0, -self._metrology_params.distance * 1e-3)),
        )[1]

        self._raw_data = []
        detector = Detector()

        for p in d.panel:
            Tb_p = (
                Tb_d * _transform(col(p.orientation).normalize(), col(p.translation))[1]
            )

            for s in p.sensor:
                Tb_s = (
                    Tb_p
                    * _transform(col(s.orientation).normalize(), col(s.translation))[1]
                )

                for a in s.asic:
                    Tb_a = (
                        Tb_s
                        * _transform(
                            col(a.orientation).normalize(), col(a.translation)
                        )[1]
                    )

                    Pb = get_projection_matrix(a.pixel_size, a.dimension)[1]

                    # The DXTBX-style metrology description consists of three
                    # vectors for each ASIC.  The origin vector locates the
                    # (0, 0)-pixel in the laboratory frame in units of mm.
                    # The second and third vectors give the directions to the
                    # pixels immediately next to (0, 0) in the fast and slow
                    # directions, respectively, in arbitrary units.
                    origin = Tb_a * Pb * col((0, 0, 1))
                    fast = Tb_a * Pb * col((0, a.dimension[0], 1)) - origin
                    slow = Tb_a * Pb * col((a.dimension[1], 0, 1)) - origin

                    # Convert vector units from meter to millimeter.  The
                    # default, SimplePxMmStrategy applies here.  XXX Due to
                    # dark subtraction, a valid pixel intensity may be
                    # negative, and this is currently not reflected by
                    # trusted_range.
                    key = (d.serial, p.serial, s.serial, a.serial)

                    panel = detector.add_panel()
                    panel.set_type("PAD")
                    panel.set_name("%d:%d:%d:%d" % key)
                    panel.set_local_frame(
                        [t * 1e3 for t in fast.elems[0:3]],
                        [t * 1e3 for t in slow.elems[0:3]],
                        [t * 1e3 for t in origin.elems[0:3]],
                    )

                    panel.set_pixel_size([t * 1e3 for t in a.pixel_size])
                    panel.set_image_size(a.dimension)
                    panel.set_trusted_range((0, a.saturation))

                    self._raw_data.append(self._tiles[key])

        return detector

    def get_raw_data(self):
        """Return a tuple of flex arrays, one for each asic"""

        return tuple(self._raw_data)

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self._metrology_params.wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        # Convert textual ISO 8601 timestamp in UTC to
        # millisecond-precision Unix epoch.
        str_min = self._metrology_params.timestamp[:16] + "UTC"
        str_sec = self._metrology_params.timestamp[17:]
        epoch = timegm(strptime(str_min, "%Y-%m-%dT%H:%M%Z")) + float(str_sec)

        return self._scan_factory.make_scan(
            image_range=(1, 1),
            exposure_times=[1e-15 * self._metrology_params.pulse_length],
            oscillation=(0, 0),
            epochs={1: epoch},
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatPYmultitile.understand(arg))



from __future__ import annotations

import pickle

import pytest

from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.model import (
    Crystal,
    CrystalFactory,
    MosaicCrystalKabsch2010,
    MosaicCrystalSauter2014,
)


@pytest.fixture()
def example_crystal():
    real_space_a = matrix.col((35.2402102454, -7.60002142787, 22.080026774))
    real_space_b = matrix.col((22.659572494, 1.47163505925, -35.6586361881))
    real_space_c = matrix.col((5.29417246554, 38.9981792999, 4.97368666613))
    return {
        "real_space_a": real_space_a,
        "real_space_b": real_space_b,
        "real_space_c": real_space_c,
        "space_group_symbol": "P 1 2/m 1",
    }


def test_crystal(example_crystal):
    c1 = Crystal(**example_crystal)

    d = c1.to_dict()
    c2 = CrystalFactory.from_dict(d)
    for direction in ("real_space_a", "real_space_b", "real_space_c"):
        assert abs(matrix.col(d[direction]) - example_crystal[direction]) <= 1e-7
    assert d["space_group_hall_symbol"] == "-P 2y"
    assert c1 == c2


def test_mosaic_crystal(example_crystal):
    c1 = MosaicCrystalKabsch2010(**example_crystal)
    c1.set_mosaicity(0.1)

    d = c1.to_dict()
    c2 = CrystalFactory.from_dict(d)
    for direction in ("real_space_a", "real_space_b", "real_space_c"):
        assert abs(matrix.col(d[direction]) - example_crystal[direction]) <= 1e-7
    assert d["space_group_hall_symbol"] == "-P 2y"
    assert d["mosaicity"] == 0.1
    assert c1 == c2


def test_crystal_with_scan_points(example_crystal):
    c1 = Crystal(**example_crystal)

    A = c1.get_A()
    c1.set_A_at_scan_points([A for i in range(5)])

    # Set the B covariance. The values are nonsense, just ensure they are
    # all different
    cov_B = flex.double(range(9 * 9)) * 1e-5
    c1.set_B_covariance(cov_B)
    cov_B.reshape(flex.grid(1, 9, 9))
    cov_B_array = flex.double(flex.grid(5, 9, 9))
    for i in range(5):
        cov_B_array[i : (i + 1), :, :] = cov_B
    c1.set_B_covariance_at_scan_points(cov_B_array)
    cov_B = c1.get_B_covariance()

    d = c1.to_dict()
    c2 = CrystalFactory.from_dict(d)
    eps = 1e-9
    for Acomp in d["A_at_scan_points"]:
        for e1, e2 in zip(A, Acomp):
            assert abs(e1 - e2) <= eps
    for covBcomp in d["B_covariance_at_scan_points"]:
        for e1, e2 in zip(cov_B, covBcomp):
            assert abs(e1 - e2) <= eps

    assert c1 == c2


@pytest.mark.parametrize(
    "crystal_class", [Crystal, MosaicCrystalKabsch2010, MosaicCrystalSauter2014]
)
def test_crystal_with_recalculated_cell(crystal_class, example_crystal):
    c1 = crystal_class(**example_crystal)
    uc = c1.get_unit_cell()
    c1.set_recalculated_unit_cell(uc)
    c1.set_recalculated_cell_parameter_sd((0.1,) * 6)
    c1.set_recalculated_cell_volume_sd(0.001)

    d = c1.to_dict()
    c2 = CrystalFactory.from_dict(d)
    c3 = pickle.loads(pickle.dumps(c1))

    for c in (c2, c3):
        assert c.get_recalculated_unit_cell() is not None
        assert c1.get_recalculated_unit_cell().is_similar_to(
            c.get_recalculated_unit_cell()
        )
        assert c1 == c
        assert c.get_recalculated_cell_parameter_sd() == (0.1, 0.1, 0.1, 0.1, 0.1, 0.1)
        assert c.get_recalculated_cell_volume_sd() == 0.001



from __future__ import annotations

import pytest

from dxtbx.model import ProfileModelFactory


def test_profile_modelling():
    grs = pytest.importorskip("dials.algorithms.profile_model.gaussian_rs")
    if not hasattr(grs, "__file__"):
        pytest.skip("test requires DIALS")
        # this may pretend to be present in Python 3 without dials actually being there

    profile1 = grs.Model(None, 3, 0.1, 0.2, deg=True)
    dictionary = profile1.to_dict()
    profile2 = ProfileModelFactory.from_dict(dictionary)
    assert profile1.sigma_b() == pytest.approx(profile2.sigma_b(), abs=1e-7)
    assert profile1.sigma_m() == pytest.approx(profile2.sigma_m(), abs=1e-7)



"""An implementation of the TIFF image reader for Rayonix images."""


from __future__ import annotations

import datetime
import struct
import sys
import time

from boost_adaptbx.boost.python import streambuf
from iotbx.detectors.mar import MARImage
from scitbx.array_family import flex

from dxtbx.ext import read_uint16
from dxtbx.format.FormatTIFF import FormatTIFF


class FormatTIFFRayonix(FormatTIFF):
    """A class for reading TIFF format Rayonix images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Rayonix TIFF format image,
        i.e. we can make sense of it. This simply checks that records which
        describe the size of the image match with the TIFF records which do
        the same."""

        width, height, depth, order, bytes = FormatTIFF.get_tiff_header(image_file)

        if len(bytes) != 4096:
            return False

        if order == FormatTIFF.LITTLE_ENDIAN:
            endian = "<"
        else:
            endian = ">"

        _I = endian + "I"

        _width = struct.unpack(_I, bytes[1024 + 80 : 1024 + 84])[0]
        _height = struct.unpack(_I, bytes[1024 + 84 : 1024 + 88])[0]
        _depth = struct.unpack(_I, bytes[1024 + 88 : 1024 + 92])[0]

        if width != _width or height != _height or depth != _depth:
            return False

        # pretty sure all MARCCD / Rayonix detectors are square...
        if width != height:
            return False

        nimages = struct.unpack(_I, bytes[1024 + 112 : 1024 + 116])[0]
        origin = struct.unpack(_I, bytes[1024 + 116 : 1024 + 120])[0]
        orientation = struct.unpack(_I, bytes[1024 + 120 : 1024 + 124])[0]
        view = struct.unpack(_I, bytes[1024 + 124 : 1024 + 128])[0]

        if nimages != 1 or origin != 0 or orientation != 0 or view != 0:
            return False

        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        width, height, depth, order, bytes = FormatTIFF.get_tiff_header(image_file)

        # comment block - where the detector serial number may (or may not) be stored
        # comments = bytes[1024+1440:1024+1440+512]

        self._header_size = 4096

        if order == FormatTIFF.LITTLE_ENDIAN:
            self._I = "<I"
            self._i = "<i"
            self._ii = "<ii"
        else:
            self._I = ">I"
            self._i = ">i"
            self._ii = ">ii"

        super().__init__(image_file, **kwargs)

    def detectorbase_start(self):
        self.detectorbase = MARImage(self._image_file)
        self.detectorbase.readHeader()

    # FIXME have implemented none of those which follow...

    def _goniometer(self):
        """Return a model for goniometer corresponding to the values stored
        in the image header. In the first instance assume this is a single
        axis and raise exception otherwise."""

        starts, ends, offset, width = self._get_rayonix_scan_angles()

        for j, starts_j in enumerate(starts):
            if j != offset:
                assert starts_j == ends[j]

        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector, which at the moment [NKS does not] insist
        that the offsets and rotations are all 0.0."""

        starts, ends, offset, width = self._get_rayonix_scan_angles()
        rotations = self._get_rayonix_detector_rotations()

        # NKS removed the assertion that two-theta offset is 0.0; support the general case
        assert starts[0] == ends[0]
        two_theta = starts[0]

        # assert that the rotations are all 0.0

        assert rotations[0] == 0.0
        assert rotations[1] == 0.0
        assert rotations[2] == 0.0

        distance = self._get_rayonix_distance()
        beam_x, beam_y = self._get_rayonix_beam_xy()
        pixel_size = self._get_rayonix_pixel_size()
        image_size = self._tiff_width, self._tiff_height
        max_trusted_value = struct.unpack(self._i, self._tiff_header_bytes[1128:1132])[
            0
        ]
        min_trusted_value = 0

        bias = int(round(self._get_rayonix_bias()))
        min_trusted_value -= bias
        max_trusted_value -= bias

        beam = beam_x * pixel_size[0], beam_y * pixel_size[1]

        return self._detector_factory.two_theta(
            "CCD",
            distance,
            beam,
            "+x",
            "-y",
            "+x",
            two_theta,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
            [],
            gain=self._get_rayonix_gain(),
        )

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = (
            struct.unpack(self._i, self._tiff_header_bytes[1932:1936])[0] * 1.0e-5
        )

        return self._beam_factory.simple(wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        exposure_time = self._get_rayonix_times()[1]
        epoch = time.mktime(self._get_rayonix_timestamp())

        starts, ends, offset, width = self._get_rayonix_scan_angles()

        osc_start = starts[offset]
        osc_range = width

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    ####################################################################
    #                                                                  #
    # Helper methods to get all of the values out of the TIFF header   #
    # - separated out to assist with code clarity                      #
    #                                                                  #
    ####################################################################

    def _get_rayonix_distance(self):
        """Look in the usual places for the detector distance, return this
        as a float in mm."""

        distance = struct.unpack(self._i, self._tiff_header_bytes[1664:1668])[0]

        if distance != 0:
            return distance * 0.001

        distance = struct.unpack(self._i, self._tiff_header_bytes[1720:1724])[0]

        if distance != 0:
            return distance * 0.001

        raise RuntimeError("cannot find distance in header")

    def _get_rayonix_beam_xy(self):
        """Get the beam x, y positions which are defined in the standard
        to be in pixels. X and Y are not defined by the documentation, so
        taking as a given that these are horizontal and vertical. N.B.
        the documentation states that the horizontal direction is fast."""

        beam_x, beam_y = struct.unpack(self._ii, self._tiff_header_bytes[1668:1676])[:2]

        return beam_x * 0.001, beam_y * 0.001

    def _get_rayonix_pixel_size(self):
        """Get the pixel sizes in mm."""

        pixel_x, pixel_y = struct.unpack(self._ii, self._tiff_header_bytes[1796:1804])[
            :2
        ]

        return pixel_x * 1.0e-6, pixel_y * 1.0e-6

    def _get_rayonix_times(self):
        """Get the integration, exposure times in seconds."""

        integration, exposure = struct.unpack(
            self._ii, self._tiff_header_bytes[1676:1684]
        )[:2]

        return integration * 0.001, exposure * 0.001

    def _get_rayonix_timestamp(self):
        """Get the image acquisition timestamp."""

        timestamp = self._tiff_header_bytes[2368:2400]

        month = int(timestamp[:2])
        day = int(timestamp[2:4])
        hour = int(timestamp[4:6])
        minute = int(timestamp[6:8])
        year = int(timestamp[8:12])
        second = int(timestamp[13:15])
        return datetime.datetime(year, month, day, hour, minute, second).timetuple()

    def _get_rayonix_scan_angles(self):
        """Get the scan angles for: twotheta, omega, chi, kappa, phi, delta,
        gamma. The exact definitions for these are somewhat poorly defined,
        though I presume that they will come from some kind of standard
        goniometer definition... Also returns the scan axis offset and
        the apparent oscillation width, as these are sometimes not properly
        recorded elsewhere."""

        if self._tiff_byte_order == FormatTIFF.LITTLE_ENDIAN:
            iiiiiii = "<iiiiiii"
        else:
            iiiiiii = ">iiiiiii"

        start_angles = struct.unpack(iiiiiii, self._tiff_header_bytes[1692:1720])[:7]

        end_angles = struct.unpack(iiiiiii, self._tiff_header_bytes[1724:1752])[:7]

        axis_offset = struct.unpack(self._i, self._tiff_header_bytes[1756:1760])[0]

        axis_range = struct.unpack(self._i, self._tiff_header_bytes[1760:1764])[0]

        starts_degrees = [s * 0.001 for s in start_angles]
        ends_degrees = [e * 0.001 for e in end_angles]

        return starts_degrees, ends_degrees, axis_offset, axis_range * 0.001

    def _get_rayonix_detector_rotations(self):
        """Get the recorded rotx, roty, rotz of the detector - which in most
        cases will probably all be 0.0."""

        if self._tiff_byte_order == FormatTIFF.LITTLE_ENDIAN:
            iii = "<iii"
        else:
            iii = ">iii"

        rot_angles = struct.unpack(iii, self._tiff_header_bytes[1764:1776])[:3]

        rot_degrees = [r * 0.001 for r in rot_angles]

        return rot_degrees

    def _get_rayonix_bias(self):
        """Get the image bias."""

        bias = struct.unpack(self._i, self._tiff_header_bytes[1804:1808])[0]

        return bias * 1.0e-3

    def _get_rayonix_gain(self, model=None):
        """Return an appropriate gain value in ADU per captured X-ray for a
        Rayonix CCD module. If the model is None (unknown) then make a guess based
        on header details"""

        # Values are based on a combination of manufacturer datasheets,
        # James Holton's list at http://bl831.als.lbl.gov/xtalsize.html and
        # empirical guesswork based on spot finding results. The accuracy should
        # not be expected to be better than 20% and indeed may be worse than that.
        # Values may refer to gain in ADU per incident photon rather than the
        # more appropriate ADU per captured photon.

        pixel_size = self._get_rayonix_pixel_size()
        image_size = self._tiff_width, self._tiff_height

        if model is None:
            model = "UNKNOWN"
            # Guesswork based on information from the header
            if image_size == (2048, 2048) and pixel_size == (0.079348, 0.079348):
                model = "165"  # 2x2 binning
            elif image_size == (3072, 3072):
                model = "225"  # 2x2 binning, 73 micron pixels. But might be 225HE?
            elif image_size == (4096, 4096):
                if pixel_size == (0.079346, 0.079346):
                    model = "325"  # 2x2 binning, 79 micron pixels. But might be 325HE?
                elif pixel_size == (0.073242, 0.073242):
                    model = "300"  # 2x2 binning, 73 micron pixels. But might be 300HE?

        model = model.upper()
        if model.startswith("165"):
            return 1.0
        if model.startswith("225HE"):
            return 2.8  # This is inferred based on the value for 300HE
        if model.startswith("225"):
            return 1.5
        if model.startswith("325HE"):
            return 2.4
        if model.startswith("325"):
            return 1.3
        if model.startswith("300HE"):
            return 2.8
        if model.startswith("300"):
            return 1.5

        # Unidentified model
        return 1.0

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        # currently have no non-little-endian machines...

        assert self._tiff_byte_order == FormatTIFF.LITTLE_ENDIAN

        bias = int(round(self._get_rayonix_bias()))

        assert len(self.get_detector()) == 1
        size = self.get_detector()[0].get_image_size()
        f = FormatTIFF.open_file(self._image_file)
        f.read(self._header_size)
        raw_data = read_uint16(streambuf(f), int(size[0] * size[1])) - bias
        image_size = self.get_detector()[0].get_image_size()
        raw_data.reshape(flex.grid(image_size[1], image_size[0]))

        return raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatTIFFRayonix.understand(arg))



from __future__ import annotations

from typing import Union

import numpy as np

from scitbx.array_family import flex

# Simple, flat array types
FlexSimpleArray = Union[
    flex.bool,
    flex.uint8,
    flex.uint16,
    flex.uint32,
    flex.uint64,
    flex.size_t,
    flex.int,
    flex.int8,
    flex.int16,
    flex.int32,
    flex.int64,
    flex.double,
    flex.float,
    flex.complex_double,
]
FlexVecArray = Union[flex.vec2_double, flex.vec3_double, flex.vec3_int]
FlexArray = Union[FlexSimpleArray, FlexVecArray]

class Scuffer:
    def __init__(self, arg0: FlexArray) -> None: ...
    @property
    def base(self) -> FlexArray: ...

def from_numpy(array: np.ndarray) -> FlexSimpleArray: ...
def mat3_from_numpy(array: np.ndarray) -> flex.mat3_double: ...
def to_numpy(array: FlexArray) -> np.ndarray: ...
def vec_from_numpy(array: np.ndarray) -> FlexVecArray: ...
